import re
from typing import Optional

import ixmp
import pytest
from message_ix import make_df

from message_ix_models import Workflow, testing
from message_ix_models.workflow import WorkflowStep, make_click_command, solve

MARK = pytest.mark.skipif(
    condition=ixmp.__version__ < "3.5",
    reason="ixmp.TimeSeries.url not available prior to ixmp 3.5.0",
)


# Functions for WorkflowSteps


def changes_a(c, s) -> None:
    """Change a scenario by modifying structure data, but not data."""
    with s.transact():
        s.add_set("technology", "test_tech")


def changes_b(c, s, value=None) -> None:
    """Change a scenario by modifying parameter data, but not structure."""
    with s.transact():
        s.add_par(
            "technical_lifetime",
            make_df(
                "technical_lifetime",
                node_loc=s.set("node")[0],
                year_vtg=s.set("year")[0],
                technology="test_tech",
                value=value,
                unit="y",
            ),
        )


class TestWorkflowStep:
    def test_call(self, test_context) -> None:
        def action(c, s):
            pass  # pragma: no cover

        ws = WorkflowStep(action=action)

        with pytest.raises(RuntimeError):
            ws(test_context, None)

    def test_repr(self) -> None:
        assert "<Step load>" == repr(WorkflowStep(None))


@pytest.fixture(scope="function")
def wf(request, test_context) -> Workflow:
    return _wf(test_context, request=request)


def _wf(
    context,
    *,
    base_url: Optional[str] = None,
    base_platform: Optional[str] = None,
    request=None,
):
    if base_url is base_platform is None:
        base_scenario = testing.bare_res(request, context, solved=False)
        base_platform = base_scenario.platform.name
        base_url = f"ixmp://{base_platform}/{base_scenario.url}"

    """A function that generates a Workflow."""
    # Create the workflow
    wf = Workflow(context)

    # Model/base is created from nothing by calling base_scenario
    wf.add_step("base", None, target=base_url)
    # Model/A is created from Model/base by calling changes_a
    wf.add_step("A", "base", changes_a)
    # Model/B is created from Model/A by calling changes_b
    wf.add_step("B", "A", changes_b, value=100.0)

    # Store extra info
    wf.graph.update({"_base_platform": base_platform})

    return wf


@MARK
def test_make_click_command(mix_models_cli) -> None:
    import click

    from message_ix_models.cli import cli_test_group
    from message_ix_models.util.click import temporary_command

    # make_click_command() runs and generates a command
    name = "make-click-command"
    cmd = make_click_command(f"{__name__}._wf", name=name, slug="test")
    assert isinstance(cmd, click.Command)

    # Add this into the hidden CLI test group
    with temporary_command(cli_test_group, cmd):
        # Invoke the command with various parameters
        for params, output in (
            (["--go", "B"], "nothing returned, workflow will continue with"),
            (["B"], "Workflow diagram written to"),
        ):
            # Command runs and exits with 0
            result = mix_models_cli.assert_exit_0(["_test", "run"] + params)
            # Expected log messages or output were printed
            assert output in result.output

        # Invalid usage
        for params, output in (
            (["--go", "C"], "Error: No step(s) matched"),
            (["--go"], "Error: No target step provided and no default for"),
            # Step changes_b() fails if changes_a() is not first run
            (["--go", "--from=[AX]", "B"], "Execute <function changes_b"),
        ):
            result = mix_models_cli.invoke(["_test", "run"] + params)
            assert 0 != result.exit_code
            assert output in result.output


@MARK
def test_workflow(caplog, request, test_context, wf) -> None:
    # Retrieve some information from the fixture
    mp = wf.graph.pop("_base_platform")

    caplog.clear()

    # "B solved" is created from "Model/B" by clone and running solve()
    # clone=True without target= raises an exception
    with pytest.raises(TypeError, match="target= must be supplied"):
        wf.add_step("B solved", "B", solve, clone=True)

    wf.add_step("B solved", "B", solve, clone=True, target="foo/bar")

    # Trigger the creation and solve of Model/B and all required precursor scenarios
    s = wf.run("B solved")

    # Scenario contains changes from the first and second step
    assert "test_tech" in set(s.set("technology"))
    assert 1 == len(s.par("technical_lifetime"))
    # Scenario was solved
    assert s.has_solution()

    # Log messages reflect workflow steps executed
    start_index = 1 if caplog.messages[0].startswith("Cull") else 0
    # This setting obtains the value R11 on some Windows GHA jobs, but is otherwise R14.
    # TODO Debug and fix.
    m = f"MESSAGEix-GLOBIOM {test_context.model.regions} YB"
    messages = [
        f"Loaded ixmp://{mp}/{m}/test_workflow#1",
        f"Step runs on ixmp://{mp}/{m}/test_workflow#1",
        "Execute <function changes_a at [^>]*>",
        f"…nothing returned, workflow will continue with {m}/test_workflow#1",
        f"Step runs on ixmp://{mp}/{m}/test_workflow#1",
        "Execute <function changes_b at [^>]*>",
        f"…nothing returned, workflow will continue with {m}/test_workflow#1",
        f"Step runs on ixmp://{mp}/{m}/test_workflow#1",
        "  with context.dest_scenario={'model': 'foo', 'scenario': 'bar'}",
        "Clone to foo/bar",
        "Execute <function solve at [^>]*>",
    ]
    for expr, message in zip(messages, caplog.messages[start_index:]):
        assert re.match(expr, message)

    assert re.match(
        rf"""'B':
- <Step changes_b\(\)>
- 'context':
  - <Context object at \w+ with \d+ keys>
- 'A':
  - <Step changes_a\(\)>
  - 'context' \(above\)
  - 'base':
    - <Step load -> {m}/test_workflow>
    - 'context' \(above\)
    - None""",
        wf.describe("B"),
    )

    # Now truncate the workflow at "Model/A"
    wf.truncate("A")

    # Description reflects that changes_a() will no longer be called
    assert re.match(
        rf"""'B':
- <Step changes_b\(\)>
- 'context':
  - <Context object at \w+ with \d+ keys>
- 'A':
  - <Step load -> {m}/test_workflow>
  - 'context' \(above\)
  - None""",
        wf.describe("B"),
    )



import logging
from dataclasses import InitVar, dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Optional, Union

from message_ix_models.util import local_data_path, package_data_path
from message_ix_models.util.config import ConfigHelper

if TYPE_CHECKING:
    from genno.core.key import KeyLike

log = logging.getLogger(__name__)


@dataclass
class Config(ConfigHelper):
    """Settings for :mod:`message_ix_models.report`.

    When initializing a new instance, the `from_file` and `_legacy` parameters are
    respected.
    """

    #: Shorthand to call :func:`use_file` on a new instance.
    from_file: InitVar[Optional[Path]] = package_data_path("report", "global.yaml")

    #: Shorthand to set :py:`legacy["use"]` on a new instance.
    _legacy: InitVar[Optional[bool]] = False

    #: Path to write reporting outputs when invoked from the command line.
    cli_output: Optional[Path] = None

    #: Configuration to be handled by :mod:`genno.config`.
    genno_config: Dict = field(default_factory=dict)

    #: Key for the Quantity or computation to report.
    key: Optional["KeyLike"] = None

    #: Directory for output.
    output_dir: Optional[Path] = field(
        default_factory=lambda: local_data_path("report")
    )

    #: :data:`True` to use an output directory based on the scenario's model name and
    #: name.
    use_scenario_path: bool = True

    #: Keyword arguments for the ‘legacy’ reporting entry point
    #: (:func:`message_data.tools.post_processing.iamc_report_hackathon.report`), plus
    #: the special key "use", which should be :data:`True` if legacy reporting is to
    #: be used.
    legacy: Dict = field(default_factory=lambda: dict(use=False, merge_hist=True))

    def __post_init__(self, from_file, _legacy) -> None:
        self.use_file(from_file)
        self.legacy.update(use=_legacy)

    def set_output_dir(self, arg: Optional[Path]) -> None:
        """Set :attr:`output_dir`, the output directory.

        The value is also stored to be passed to :mod:`genno` as the "output_dir"
        configuration key.
        """
        if arg:
            self.output_dir = arg.expanduser()

        self.genno_config["output_dir"] = self.output_dir

    def use_file(self, file_path: Union[str, Path, None]) -> None:
        """Use genno configuration from a (YAML) file at `file_path`.

        See :mod:`genno.config` for the format of these files. The path is stored at
        :py:`.genno_config["path"]`, where it is picked up by genno's configuration
        mechanism.

        Parameters
        ----------
        file_path : PathLike, *optional*
            This may be:

            1. The complete path to any existing file.
            2. A stem like "global" or "other". This is interpreted as referring to a
               file named, for instance, :file:`global.yaml`.
            3. A partial path like "project/report.yaml". This or (2) is interpreted
               as referring to a file within :file:`MESSAGE_MODELS_PATH/data/report/`;
               that is, a file packaged and distributed with :mod:`message_ix_models`.
        """
        if file_path is None:
            return
        try:
            path = next(
                filter(
                    Path.exists,
                    (
                        Path(file_path),
                        # Path doesn't exist; treat it as a stem in the metadata dir
                        package_data_path("report", file_path).with_suffix(".yaml"),
                    ),
                )
            )
        except StopIteration:
            raise FileNotFoundError(f"Reporting configuration in '{file_path}(.yaml)'")

        # Store for genno to handle
        self.genno_config["path"] = path

    def mkdir(self) -> None:
        """Ensure the :attr:`output_dir` exists."""
        if self.output_dir:
            self.output_dir.mkdir(exist_ok=True, parents=True)



from . import operator

_warned = False


def __getattr__(name):
    global _warned
    if not _warned:
        from warnings import warn

        warn(
            f"Importing from {__name__} is deprecated and will be removed on or after "
            "2024-02-04; use message_ix_models.report.operator instead.",
            DeprecationWarning,
            2,
        )
        _warned = True

    return getattr(operator, name)



import logging
from typing import Dict, Iterable, Optional, Union

import pandas as pd
from dask.core import quote
from genno import Key, Quantity
from genno.compat.pyam.util import collapse as genno_collapse
from genno.core.key import single_key
from iam_units import registry
from message_ix import Reporter
from sdmx.model.v21 import Code

log = logging.getLogger(__name__)


#: Replacements used in :meth:`collapse`.
#: These are applied using :meth:`pandas.DataFrame.replace` with ``regex=True``; see the
#: documentation of that method.
#:
#: - Applied to whole strings along each dimension.
#: - These columns have :meth:`str.title` applied before these replacements.
REPLACE_DIMS: Dict[str, Dict[str, str]] = {
    "c": {
        # in land_out, for CH4 emissions from GLOBIOM
        "Agri_Ch4": "GLOBIOM|Emissions|CH4 Emissions Total",
    },
    "l": {
        # FIXME this is probably not generally applicable and should be removed
        "Final Energy": "Final Energy|Residential",
    },
    "t": dict(),
}

#: Replacements used in :meth:`collapse` after the 'variable' column is assembled.
#: These are applied using :meth:`pandas.DataFrame.replace` with ``regex=True``; see
#: the documentation of that method. For documentation of regular expressions, see
#: https://docs.python.org/3/library/re.html and https://regex101.com.
#:
#: .. todo:: These may be particular or idiosyncratic to a single "template". The
#:    strings used to collapse multiple conceptual dimensions into the IAMC "variable"
#:    column are known to vary in poorly-documented ways across these templates.
#:
#:    This setting is currently applied universally. To improve, specify a different
#:    mapping with the replacements needed for each individual template, and load the
#:    correct one when reporting scenarios to that template.
REPLACE_VARS = {
    # Secondary energy: remove duplicate "Solids"
    r"(Secondary Energy\|Solids)\|Solids": r"\1",
    # CH4 emissions from MESSAGE technologies
    r"(Emissions\|CH4)\|Fugitive": r"\1|Energy|Supply|Fugitive",
    # CH4 emissions from GLOBIOM
    r"(Emissions\|CH4)\|((Gases|Liquids|Solids|Elec|Heat)(.*))": (
        r"\1|Energy|Supply|\3|Fugitive\4"
    ),
    r"^(land_out CH4.*\|)Awm": r"\1Manure Management",
    r"^land_out CH4\|Emissions\|Ch4\|Land Use\|Agriculture\|": (
        "Emissions|CH4|AFOLU|Agriculture|Livestock|"
    ),
    # Strip internal prefix
    r"^land_out CH4\|": "",
    # Prices
    r"Residential\|(Biomass|Coal)": r"Residential|Solids|\1",
    r"Residential\|Gas": "Residential|Gases|Natural Gas",
    r"Import Energy\|Lng": "Primary Energy|Gas",
    r"Import Energy\|Coal": "Primary Energy|Coal",
    r"Import Energy\|Oil": "Primary Energy|Oil",
    r"Import Energy\|(Liquids\|(Biomass|Oil))": r"Secondary Energy|\1",
    r"Import Energy\|Lh2": "Secondary Energy|Hydrogen",
}


def as_quantity(info: Union[dict, float, str]) -> Quantity:
    """Convert values from a :class:`dict` to Quantity.

    .. todo:: move upstream, to :mod:`genno`.
    """
    if isinstance(info, str):
        q = registry.Quantity(info)
        return Quantity(q.magnitude, units=q.units)
    elif isinstance(info, float):
        return Quantity(info)
    elif isinstance(info, dict):
        data = info.copy()
        dim = data.pop("_dim")
        unit = data.pop("_unit")
        return Quantity(pd.Series(data).rename_axis(dim), units=unit)
    else:
        raise TypeError(type(info))


def collapse(df: pd.DataFrame, var=[]) -> pd.DataFrame:
    """Callback for the `collapse` argument to :meth:`~.Reporter.convert_pyam`.

    Replacements from :data:`REPLACE_DIMS` and :data:`REPLACE_VARS` are applied.
    The dimensions listed in the `var` arguments are automatically dropped from the
    returned :class:`pyam.IamDataFrame`. If ``var[0]`` contains the word "emissions",
    then :meth:`collapse_gwp_info` is invoked.

    Adapted from :func:`genno.compat.pyam.collapse`.

    Parameters
    ----------
    var : list of str, *optional*
        Strings or dimensions to concatenate to the 'Variable' column. The first of
        these is usually a string value used to populate the column. These are joined
        using the pipe ('|') character.

    See also
    --------
    REPLACE_DIMS
    REPLACE_VARS
    collapse_gwp_info
    test_collapse
    """
    # Convert some dimension labels to title-case strings
    for dim in filter(lambda d: d in df.columns, "clt"):
        df[dim] = df[dim].astype(str).str.title()

    if "l" in df.columns:
        # Level: to title case, add the word 'energy'
        df["l"] = df["l"] + " Energy"

    if len(var) and "emissions" in var[0].lower():
        log.info(f"Collapse GWP info for {var[0]}")
        df, var = collapse_gwp_info(df, var)

    # - Apply replacements to individual dimensions.
    # - Use the genno built-in to assemble the variable column.
    # - Apply replacements to assembled columns.
    return (
        df.replace(REPLACE_DIMS, regex=True)
        .pipe(genno_collapse, columns=dict(variable=var))
        .replace(dict(variable=REPLACE_VARS), regex=True)
    )


def collapse_gwp_info(df, var):
    """:meth:`collapse` helper for emissions data with GWP dimensions.

    The dimensions 'e equivalent', and 'gwp metric' dimensions are combined
    with the 'e' dimension, using a format like::

        '{e} ({e equivalent}-equivalent, {GWP metric} metric)'

    For example::

        'SF6 (CO2-equivalent, AR5 metric)'
    """
    # Check that *df* contains the necessary columns
    cols = ["e equivalent", "gwp metric"]
    missing = set(["e"] + cols) - set(df.columns)
    if len(missing):
        log.warning(f"…skip; {missing} not in columns {list(df.columns)}")
        return df, var

    # Format the column with original emissions species
    df["e"] = (
        df["e"]
        + " ("
        + df["e equivalent"]
        + "-equivalent, "
        + df["gwp metric"]
        + " metric)"
    )

    # Remove columns from further processing
    [var.remove(c) for c in cols]
    return df.drop(cols, axis=1), var


def copy_ts(rep: Reporter, other: str, filters: Optional[dict]) -> Key:
    """Prepare `rep` to copy time series data from `other` to `scenario`.

    Parameters
    ----------
    other_url : str
       URL of the other scenario from which to copy time series data.
    filters : dict, *optional*
       Filters; passed via :func:`.store_ts` to :meth:`ixmp.TimeSeries.timeseries`.

    Returns
    -------
    str
        Key for the copy operation.
    """

    # A unique ID for this copy operation, to avoid collision if copy_ts() used multiple
    # times
    _id = f"{hash(other + repr(filters)):x}"

    k1 = rep.add("from_url", f"scenario {_id}", quote(other))
    k2 = rep.add("get_ts", f"ts data {_id}", k1, filters)
    return single_key(rep.add("store_ts", f"copy ts {_id}", "scenario", k2))


def add_replacements(dim: str, codes: Iterable[Code]) -> None:
    """Update :data:`REPLACE_DIMS` for dimension `dim` with values from `codes`."""
    for code in codes:
        try:
            label = str(code.get_annotation(id="report").text)
        except KeyError:
            pass
        else:
            REPLACE_DIMS[dim][f"{code.id.title()}$"] = label



import logging

import pytest
from ixmp.testing import assert_logs
from message_ix import make_df
from message_ix.testing import make_dantzig

from message_ix_models import ScenarioInfo
from message_ix_models.model.build import apply_spec


@pytest.fixture
def scenario(test_context):
    mp = test_context.get_platform()
    yield make_dantzig(mp)


@pytest.fixture(scope="function")
def spec():
    yield dict(add=ScenarioInfo(), require=ScenarioInfo(), remove=ScenarioInfo())


def test_apply_spec0(caplog, scenario, spec):
    """Require missing element raises ValueError."""
    spec["require"].set["node"].append("vienna")

    with pytest.raises(ValueError):
        apply_spec(scenario, spec)

    assert (
        "message_ix_models.model.build",
        logging.ERROR,
        "  1 elements not found: ['vienna']",
    ) in caplog.record_tuples


def test_apply_spec1(caplog, scenario, spec):
    """Add data using the data= argument."""

    def add_data_func(scenario, dry_run):
        return dict(
            demand=make_df(
                "demand",
                commodity="cases",
                level="consumption",
                node="chicago",
                time="year",
                unit="case",
                value=301.0,
                year=1963,
            )
        )

    apply_spec(scenario, spec, data=add_data_func, dry_run=True)

    # Messages are logged about additions
    assert_logs(caplog, "1 rows in 'demand'")
    # …but no change because `dry_run`
    assert not any(301.0 == scenario.par("demand")["value"])

    caplog.clear()

    apply_spec(scenario, spec, data=add_data_func)

    # Messages are logged about additions
    assert_logs(caplog, "1 rows in 'demand'")
    # Value was actually changed
    assert 1 == sum(301.0 == scenario.par("demand")["value"])


def test_apply_spec2(caplog, scenario, spec):
    """Remove an element, with fast=True."""
    spec["remove"].set["node"] = ["new-york", "not-a-node"]

    apply_spec(scenario, spec, fast=True)

    # Messages are logged about removals
    assert_logs(
        caplog,
        (
            "Remove 'new-york' from set 'node'",
            "Remove 'not-a-node' from set 'node'",
            "  …not found",
        ),
    )


def test_apply_spec3(caplog, scenario, spec):
    """Actually remove data."""
    spec["remove"].set["node"] = ["new-york"]

    apply_spec(scenario, spec)

    # Messages are logged about removals
    assert_logs(
        caplog,
        (
            "Remove data with node='new-york'",
            "  1 rows in 'demand'",
            "  2 rows in 'output'",
            "  3 rows total",
        ),
    )



import logging
import shutil
import sys

import pytest
from message_ix import Scenario

from message_ix_models.model import snapshot
from message_ix_models.testing import GHA
from message_ix_models.util import package_data_path
from message_ix_models.util.pooch import SOURCE

log = logging.getLogger(__name__)


@pytest.fixture
def unpacked_snapshot_data(test_context, request):
    """Already-unpacked data for a snapshot.

    This copies the .csv.gz files from message_ix_models/tests/data/… to the directory
    where they *would* be unpacked by .model.snapshot._unpack. This causes the code to
    skip unpacking them, which can be very slow.
    """
    snapshot_id = request.getfixturevalue("snapshot_id")
    assert 0 == snapshot_id, f"No unpacked data for snapshot {snapshot_id}"

    dest = test_context.get_cache_path("MESSAGEix-GLOBIOM_1.1_R11_no-policy_baseline")
    log.debug(f"{dest = }")

    snapshot_data_path = package_data_path(
        "test", "MESSAGEix-GLOBIOM_1.1_R11_no-policy_baseline"
    )
    log.debug(f"{snapshot_data_path = }")

    shutil.copytree(snapshot_data_path, dest, dirs_exist_ok=True)


@snapshot.load.minimum_version
@pytest.mark.skipif(
    condition=GHA and sys.platform in ("darwin", "win32"), reason="Slow."
)
@pytest.mark.usefixtures("unpacked_snapshot_data")
@pytest.mark.parametrize(
    "snapshot_id", [int(k.split("-")[1]) for k in SOURCE if k.startswith("snapshot")]
)
def test_load(test_context, snapshot_id):
    mp = test_context.get_platform()
    base = Scenario(mp, model="MODEL", scenario="baseline", version="new")

    snapshot.load(base, snapshot_id)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full list see
# the documentation: https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
from pathlib import Path
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    import sphinx.application

# -- Project information ---------------------------------------------------------------

project = "message-ix-models"
copyright = "2020–2024, IIASA Energy, Climate, and Environment (ECE) Program"
author = "IIASA Energy, Climate, and Environment (ECE) Program"


# -- General configuration -------------------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions coming
# with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    # "ixmp.util.sphinx_linkcode_github",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.viewcode",
    "sphinxcontrib.bibtex",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_template"]

# List of patterns, relative to source directory, that match files and directories to
# ignore when looking for source files. This pattern also affects html_static_path and
# html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

nitpicky = True
nitpick_ignore_regex = {
    # These occur because there is no .. py:module:: directive for the *top-level*
    # module or package in the respective documentation and inventories.
    # TODO Remove once the respective docs are fixed
    ("py:mod", "ixmp"),
    ("py:mod", "message_ix"),
    ("py:mod", "message_data"),
    # iam-units has no Sphinx docs
    ("py:.*", "iam_units.*"),
    # These are a consequence of autosummary-class.rst
    ("py:(obj|meth)", r".*\.Test.*\.test_.*"),
}

rst_prolog = """
.. role:: py(code)
   :language: python

.. |Annotation| replace:: :class:`~sdmx.model.common.Annotation`
.. |Code| replace:: :class:`~sdmx.model.common.Code`
.. |Codelist| replace:: :class:`~sdmx.model.common.Codelist`
.. |Platform| replace:: :class:`~ixmp.Platform`
.. |Scenario| replace:: :class:`~message_ix.Scenario`

.. |n| replace:: :math:`n`
.. |y| replace:: :math:`y`
.. |y0| replace:: :math:`y_0`
"""


def setup(app: "sphinx.application.Sphinx") -> None:
    """Copied from pytest's conf.py to enable intersphinx references to these."""
    app.add_crossref_type(
        "fixture",
        "fixture",
        objname="built-in fixture",
        indextemplate="pair: %s; fixture",
    )


# -- Options for HTML output -----------------------------------------------------------

# A list of CSS files.
html_css_files = ["custom.css"]

html_favicon = "_static/favicon.svg"

# The name of an image file (relative to this directory) to place at the top of the
# sidebar.
html_logo = "_static/combined-logo-white.png"

# Add any paths that contain custom static files (such as style sheets) here, relative
# to this directory. They are copied after the builtin static files, so a file named
# "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# The theme to use for HTML and HTML Help pages.  See the documentation for a list of
# builtin themes.
html_theme = "sphinx_rtd_theme"

# -- Options for sphinx.ext.autosummary ------------------------------------------------

autosummary_generate = True

# -- Options for sphinx.ext.extlinks ---------------------------------------------------

extlinks = {
    "issue": ("https://github.com/iiasa/message-ix-models/issue/%s", "GH #%s"),
    "pull": ("https://github.com/iiasa/message-ix-models/pull/%s", "PR #%s"),
}

# -- Options for sphinx.ext.intersphinx ------------------------------------------------


def local_inv(name: str, *parts: str) -> Optional[str]:
    """Construct the path to a local intersphinx inventory."""
    if 0 == len(parts):
        parts = ("doc", "_build", "html")

    from importlib.util import find_spec

    spec = find_spec(name)
    if spec and spec.origin:
        return str(Path(spec.origin).parents[1].joinpath(*parts, "objects.inv"))


# For message-data, see: https://docs.readthedocs.io/en/stable/guides
# /intersphinx.html#intersphinx-with-private-projects
_token = os.environ.get("RTD_TOKEN_MESSAGE_DATA", "")

intersphinx_mapping = {
    "click": ("https://click.palletsprojects.com/en/8.1.x/", None),
    "genno": ("https://genno.readthedocs.io/en/stable", None),
    "ixmp": ("https://docs.messageix.org/projects/ixmp/en/latest/", None),
    "message-ix": ("https://docs.messageix.org/en/latest/", None),
    "m-data": (
        f"https://{_token}:@docs.messageix.org/projects/models-internal/en/latest/",
        # Use a local copy of objects.inv, if the user has one
        (local_inv("message_data"), None),
    ),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "pint": ("https://pint.readthedocs.io/en/stable/", None),
    "platformdirs": ("https://platformdirs.readthedocs.io/en/latest", None),
    "pooch": ("https://www.fatiando.org/pooch/latest/", None),
    "pytest": ("https://docs.pytest.org/en/stable/", None),
    "python": ("https://docs.python.org/3/", None),
    "sdmx": ("https://sdmx1.readthedocs.io/en/stable/", None),
}

# -- Options for sphinx.ext.linkcode / ixmp.util.sphinx_linkcode_github ----------------

linkcode_github_repo_slug = "iiasa/message-ix-models"

# -- Options for sphinx.ext.napoleon ---------------------------------------------------

napoleon_preprocess_types = True
napoleon_type_aliases = {
    "Code": ":class:`~sdmx.model.common.Code`",
    "Path": ":class:`~pathlib.Path`",
    "PathLike": ":class:`os.PathLike`",
}

# -- Options for sphinx.ext.todo -------------------------------------------------------

todo_include_todos = True

# -- Options for sphinxcontrib.bibtex --------------------------------------------------

bibtex_bibfiles = ["main.bib"]



"""
This script aggregates the global gridded data to any
scale and also adjust unit conversions. The following
script specifically aggregates global gridded hydrological
data onto the basin
 mapping used in the nexus module.
"""

import glob

#  Import packages
import os
from datetime import datetime as dt

import dask
import numpy as np
import xarray as xr

start = dt.now()


# variable, for detailed symbols, refer to ISIMIP2b documentation
variables = [
    "qtot",  # total runoff
    "dis",  # discharge
    "qg",  # groundwater runoff
    "qr",
]  # groudnwater recharge
var = "qr"

isimip = "3b"

data = "future"  # else future
climmodels = ["gfdl-esm2m", "hadgem2-es", "ipsl-cm5a-lr", "miroc5"]

for cl in climmodels:
    # climate model
    if isimip == "2b":
        climmodels = ["gfdl-esm2m", "hadgem2-es", "ipsl-cm5a-lr", "miroc5"]
        climmodel = "gfdl-esm2m"
        # climate forcing
        scenarios = ["rcp26", "rcp60"]
        scen = "rcp26"
        wd1 = os.path.join("p:", "ene.model", "NEST", "hydrological_data_agg") + os.sep
        wd = (
            os.path.join("p:", "watxene", "ISIMIP", "ISIMIP2b", "output", "LPJmL")
            + os.sep
        )
        wd2 = (
            os.path.join("p:", "ene.model", "NEST", "hydrology", "processed_nc4")
            + os.sep
        )
    else:
        climmodels = [
            "gfdl-esm4",
            "ipsl-cm6a-lr",
            "mpi-esm1-2-hr",
            "mri-esm2-0",
            "ukesm1-0-ll",
        ]
        scenarios = ["ssp126", "ssp370", "ssp585"]
        scen = "ssp126"
        wd1 = os.path.join("p:", "ene.model", "NEST", "hydrological_data_agg")
        wd = (
            os.path.join(
                "p:",
                "watxene",
                "ISIMIP",
                "ISIMIP3b",
                "CWatM_results",
                f"{cl}",
                f"{data}",
            )
            + os.sep
        )
        wd2 = (
            os.path.join("p:", "ene.model", "NEST", "hydrology", "processed_nc4")
            + os.sep
        )

    # Define if monthly aggregation is required
    monthlyscale = True
    # define qauantile for statistical aggregation
    quant = 0.1
    # define if multi climate models mean
    multimodelensemble = True
    # deinfe lat and long chunk for reducing computational load
    latchunk = 120
    lonchunk = 640
    # Define if use all touched raster
    all_touched = True

    if var == "dis":
        # define a spatial method to aggregate
        spatialmethod = "meansd"
    else:
        # define a spatial method to aggregate
        spatialmethod = "sum"

    """
    The hydrological data can be accessed in watxene p drive. For accessing
    particular drive, seek permission from Edward Byers (byers@iiasa.ac.at)
    The files should be copied on to local drive
    """

    # Open raster area file
    # The file landareamaskmap0.nc can be found under
    # P:\ene.model\NEST\delineation\data\delineated_basins_new
    area = xr.open_dataarray(wd1 + "landareamaskmap0.nc")

    # TO AVOID ERROR WHEN OPENING AND SLICING INPUT DATA - CHECK!
    dask.config.set({"array.slicing.split-large-chunks": False})

    if data == "historical":
        hydro_data = wd + f"*{cl}*{var}*monthly*.nc"
    elif data == "future":
        hydro_data = wd + f"*{cl}*{scen}*{var}*monthly*.nc"

    files = glob.glob(hydro_data)
    if var != "qr":
        # Open hydrological data as a combined dataset
        da = xr.open_mfdataset(files)
    else:
        # Open hydrological data as a combined dataset
        da = xr.open_mfdataset(files)

        # da.to_netcdf(wd+f'\{var}_memean_{scen}.nc')

    if monthlyscale:
        years = np.arange(2010, 2105, 5)
    else:
        years = np.arange(2010, 2105, 5)

    # ds = xr.open_dataset(wd+f'{var}_memean_{scen}.nc')
    # Resample daily data to monthly by summing daily values

    # if var != "qr"\:
    #     da = da.resample(time="M").mean()
    #     # Calculate roolling average of 3 months to make the data consistent
    #     da = da.rolling(time=3, min_periods=1).mean()

    # da = da.fillna(0)

    # %%
    if monthlyscale:
        if var == "dis":
            # # Converts the discharge  into km3/year
            da = da * 0.031556952
            da["dis"] = da.dis.chunk(
                {"lat": latchunk, "lon": lonchunk, "time": len(da.dis.time)}
            )
            # slice is required to reduce the dataset.
            # da = da.sel(time=slice("2010-01-01", "2050-12-31"))

            da.dis.attrs["unit"] = "km3/year"
            # chunking reduces computational burden
            da["dis"] = da.dis.chunk(
                {"lat": latchunk, "lon": lonchunk, "time": len(da.dis.time)}
            )
            # saving the file here and reloading reduces disk space
            da.to_netcdf(wd2 + f"{var}_monthly__{cl}_{scen}.nc")
        elif var == "qtot":
            # da["qtot"] = da.qtot.chunk(
            #     {"lat": latchunk, "lon": lonchunk, "time": len(da.qtot.time)}
            # )
            # Filter 5 year interval monthly time step
            # da = da.sel(time=da.time.dt.year.isin(years))
            # 1kg/m2/sec = 86400 mm/day
            # 86400 mm/day X  Area (mm2) = 86400 mm3/day
            # 86400 mm3/day = 86400 X  1000000 3.65 e-16 km3/year
            da = da * 86400 * area * 3.65e-16 * 1000000

            # Resample daily data to monthly (by summing daily values)
            # da = da.resample(time="M").mean()

            # da.rolling(time=12, center=True).construct(
            #     'tmp').quantile(.1, dim='tmp', skipna=False).dropna(
            #         'time')

            # da = da.rolling(time=20, min_periods=1).mean()
            # da["qtot"] = da.qtot.chunk(
            #     {"lat": latchunk, "lon": lonchunk, "time": len(da.qtot.time)}
            # )

            da.to_netcdf(wd2 + f"{var}_monthly_{cl}_{scen}_{data}.nc")
        elif var == "qr":
            # Filter 5 year interval monthly time step
            # da = da.sel(time=da.time.dt.year.isin(years))
            # 1kg/m2/sec = 86400 mm/day
            # 86400 mm/day X  Area (mm2) = 86400 mm3/day
            # 86400 mm3/day = 86400 X  1000000 3.65 e-16 km3/year
            da = da * 86400 * area * 3.65e-16 * 1000000
            da.qr.attrs["unit"] = "km3/year"
            # da["qr"] = da.qr.chunk(
            #     {"lat": latchunk, "lon": lonchunk, "time": len(da.qr.time)}
            # )

            da.to_netcdf(wd2 + f"{var}_monthly_{cl}_{scen}_{data}.nc")

    else:
        if var == "dis":
            da["dis"] = da.dis.chunk(
                {"lat": latchunk, "lon": lonchunk, "time": len(da.dis.time)}
            )
            # Resample daily data to monthly (by summing daily values)
            da = da.resample(time="Y").mean()

            da["dis"] = da.dis.chunk(
                {"lat": latchunk, "lon": lonchunk, "time": len(da.dis.time)}
            )
            # Take 20 year rolling average to make the time scale cosnistent
            da = da.rolling(time=20).mean()

            # Converts the discharge  into km3/year
            da = da * 0.031556952
            da.dis.attrs["unit"] = "km3/year"
            # Long term Mean annual  discharge
            davg = da.resample(time="30Y", loffset="4Y").mean()
            davg["dis"] = davg.dis.chunk(
                {"lat": latchunk, "lon": lonchunk, "time": len(davg.dis.time)}
            )
            davg.to_netcdf(wd2 + f"{var}_90Y_avg_5y__{climmodel}_{scen}_temp_agg.nc")

            # Now resample to an average value for each 5-year block, and
            # offset by 2 years so that the value is centred to start 2101
            da = da.resample(time="5Y", loffset="4Y").mean()

            da.to_netcdf(wd2 + f"{var}_5y__{climmodel}_{scen}_temp_agg.nc")
            # read in saved data again to further process
            da = xr.open_dataset(wd2 + f"{var}_5y__{climmodel}_{scen}_temp_agg.nc")

        elif var == "qtot":
            da["qtot"] = da.qtot.chunk(
                {"lat": latchunk, "lon": lonchunk, "time": len(da.qtot.time)}
            )
            # Resample daily data to monthly (by summing daily values)
            da = da.resample(time="Y").mean()
            da["qtot"] = da.qtot.chunk(
                {"lat": latchunk, "lon": lonchunk, "time": len(da.qtot.time)}
            )
            da = da.rolling(time=20, min_periods=1).mean()
            da["qtot"] = da.qtot.chunk(
                {"lat": latchunk, "lon": lonchunk, "time": len(da.qtot.time)}
            )
            # Converts to total runoff per gs into km3/year
            da = da * 86400 * area * 3.65e-16 * 1000000
            da.qtot.attrs["unit"] = "km3/year"
            da["qtot"] = da.qtot.chunk(
                {"lat": latchunk, "lon": lonchunk, "time": len(da.qtot.time)}
            )

            # Calculate roolling average of 20 years to make the data consistent
            # da = da.rolling(time = 20).mean()
            # Now resample to an average value for each 5-year block
            da = da.resample(time="5Y", loffset="4Y").mean()
            da.to_netcdf(wd2 + f"{var}_monthly__{climmodel}_{scen}_temp_agg.nc")
            # read in saved data again to further process
            # da = xr.\
            #    open_dataset( wd2+f'\output\{var}_5y__{climmodel}_{scen}_temp_agg.nc')
            # da['qtot'] = da.qtot.chunk({'lat': latchunk, 'lon': lonchunk,\
            #                                    'time': len(da.qtot.time)})

        elif var == "qr":
            da["qr"] = da.qr.chunk(
                {"lat": latchunk, "lon": lonchunk, "time": len(da.qr.time)}
            )
            # Resample daily data to monthly (by summing daily values)
            da = da.resample(time="Y").mean()
            da["qr"] = da.qr.chunk(
                {"lat": latchunk, "lon": lonchunk, "time": len(da.qr.time)}
            )
            da = da.rolling(time=20, min_periods=1).mean()
            da["qr"] = da.qr.chunk(
                {"lat": latchunk, "lon": lonchunk, "time": len(da.qr.time)}
            )
            # TODO check unit
            # 1kg/m2/sec = 86400 mm/day
            # 86400 mm/day X  Area (mm2) = 86400 mm3/day
            # 86400 mm3/day = 86400 X  1000000 3.65 e-16 km3/year
            da = da * 86400 * area * 3.65e-16 * 1000000
            da.qr.attrs["unit"] = "km3/year"
            da["qr"] = da.qr.chunk(
                {"lat": latchunk, "lon": lonchunk, "time": len(da.qr.time)}
            )

            # Calculate roolling average of 20 years to make the data consistent
            # da = da.rolling(time = 3).mean()
            # Now resample to an average value for each 5-year block
            da = da.resample(time="5Y", loffset="4Y").mean()

            da.to_netcdf(wd2 + f"{var}_5y__{climmodel}_{scen}.nc")



"""Tools for IEA (Extended) World Energy Balance (WEB) data."""

import logging
import zipfile
from copy import copy
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Iterable, Optional

import dask.dataframe as dd
import pandas as pd
from genno import Quantity
from genno.core.key import single_key
from platformdirs import user_cache_path

from message_ix_models.tools.exo_data import ExoDataSource, register_source
from message_ix_models.util import (
    HAS_MESSAGE_DATA,
    cached,
    package_data_path,
    private_data_path,
)
from message_ix_models.util._logging import silence_log

if TYPE_CHECKING:
    import os

    import genno

log = logging.getLogger(__name__)

#: Dimensions of the data.
DIMS = ["COUNTRY", "PRODUCT", "TIME", "FLOW", "MEASURE"]

#: Mapping from (provider, year, time stamp) → set of file name(s) containing data.
FILES = {
    ("IEA", "2023"): ("WBIG1.zip", "WBIG2.zip"),  # Timestamped 20230726T0014
    ("OECD", "2021"): ("cac5fa90-en.zip",),  # Timestamped 20211119T1000
    ("OECD", "2022"): ("372f7e29-en.zip",),  # Timestamped 20230406T1000
    ("OECD", "2023"): ("8624f431-en.zip",),  # Timestamped 20231012T1000
}


@register_source
class IEA_EWEB(ExoDataSource):
    """Provider of exogenous data from the IEA Extended World Energy Balances.

    To use data from this source, call :func:`.exo_data.prepare_computer` with the
    :py:`source_kw`:

    - "provider": Either "IEA" or "OECD". See :data:`.FILES`.
    - "edition": one of "2021", "2022", or "2023". See :data:`.FILES`.
    - "product": :class:`str` or :class:`list` of :class:`str.
    - "flow": :class:`str` or :class:`list` of :class:`str.

    The returned data have the extra dimensions "product" and "flow", and are not
    aggregated by year.

    Example
    -------
    >>> keys = prepare_computer(
    ...     context,
    ...     computer,
    ...     source="IEA_EWEB",
    ...     source_kw=dict(
    ...         provider="OECD", edition="2022", product="CHARCOAL", flow="RESIDENT"
    ...     ),
    ... )
    >>> result = computer.get(keys[0])
    """

    id = "IEA_EWEB"

    extra_dims = ("product", "flow")

    def __init__(self, source, source_kw):
        """Initialize the data source."""
        if source != self.id:
            raise ValueError(source)

        _kw = copy(source_kw)

        provider = _kw.pop("provider", None)
        edition = _kw.pop("edition", None)
        assert (provider, edition) in FILES
        self.load_kw = dict(provider=provider, edition=edition)

        self.indexers = dict(MEASURE="TJ")
        if product := _kw.pop("product", None):
            self.indexers.update(product=product)
        if flow := _kw.pop("flow", None):
            self.indexers.update(flow=flow)

        if len(_kw):
            raise ValueError(_kw)

    def __call__(self):
        """Load and process the data."""
        # - Load the data.
        # - Convert to pd.Series, then genno.Quantity.
        # - Map dimensions.
        # - Apply `indexers` to select.
        return (
            Quantity(load_data(**self.load_kw).set_index(DIMS)["Value"], units="TJ")
            .rename({"COUNTRY": "n", "TIME": "y", "FLOW": "flow", "PRODUCT": "product"})
            .sel(self.indexers, drop=True)
        )

    def transform(self, c: "genno.Computer", base_key: "genno.Key") -> "genno.Key":
        """Aggregate only; do not interpolate on "y"."""
        return single_key(
            c.add(base_key + "1", "aggregate", base_key, "n::groups", keep=False)
        )


def fwf_to_csv(path: Path, progress: bool = False) -> Path:  # pragma: no cover
    """Convert the IEA fixed-width file format to CSV.

    This appears to operate at about 900k lines / second, about 1 minute for the IEA
    2023 .TXT files. This is faster than doing full pandas I/O, which takes 5–10 minutes
    depending on formats.
    """
    import io
    import re

    # Output path
    path_out = path.with_suffix(".csv")
    if path_out.exists() and path_out.stat().st_mtime > path.stat().st_mtime:
        log.info(f"Skip conversion; file exists and is newer than source: {path_out}")
        return path_out

    # Input and output buffers; read the entire file into memory immediately
    file_in = io.BytesIO(path.read_bytes())
    file_out = io.BytesIO()

    # Regular expression to split lines
    expr = re.compile(b"  +")

    if progress:
        from tqdm import tqdm

        iterator: Iterable[bytes] = tqdm(file_in, desc=f"{path} → {path_out}")
    else:
        iterator = file_in

    # Convert to CSV
    for line in iterator:
        file_out.write(b",".join(expr.split(line)))

    # Write to file
    path_out.write_bytes(file_out.getbuffer())

    return path_out


def unpack_zip(path: Path) -> Path:
    """Unpack a ZIP archive."""
    cache_dir = user_cache_path("message-ix-models", ensure_exists=True).joinpath("iea")

    log.info(f"Decompress {path} to {cache_dir}")
    with zipfile.ZipFile(path) as zf:
        members = zf.infolist()
        assert 1 == len(members)
        zi = members[0]

        # Candidate path for the extracted file
        target = cache_dir.joinpath(zi.filename)
        if target.exists() and target.stat().st_size >= zi.file_size:
            log.info(f"Skip extraction of {target}")
            return target
        else:
            return Path(zf.extract(members[0], path=cache_dir))


@cached
def iea_web_data_for_query(
    base_path: Path, *filenames: str, query_expr: str
) -> pd.DataFrame:
    """Load data from `base_path` / `filenames` in IEA WEB formats."""
    # Filenames to pass to dask.dataframe
    names_to_read = []

    # Iterate over origin filenames
    for filename in filenames:
        path = base_path.joinpath(filename)

        if path.suffix == ".zip":
            path = unpack_zip(path)

        if path.suffix == ".TXT":  # pragma: no cover
            names_to_read.append(fwf_to_csv(path, progress=True))
            args: Dict[str, Any] = dict(header=None, names=DIMS + ["Value"])
        else:
            names_to_read.append(path)
            args = dict(header=0, usecols=DIMS + ["Value"])

    with silence_log("fsspec.local"):
        ddf = dd.read_csv(names_to_read, engine="pyarrow", **args)
        ddf = ddf[ddf["MEASURE"] == "TJ"]
        # NB compute() must precede query(), else "ValueError: The columns in the
        # computed data do not match the columns in the provided metadata" occurs with
        # the CSV-formatted data.
        result = ddf.compute().query(query_expr).dropna(subset=["Value"])

    log.info(f"{len(result)} observations")
    return result


def load_data(
    provider: str,
    edition: str,
    query_expr="MEASURE == 'TJ' and TIME >= 1980",
    path: Optional[Path] = None,
) -> pd.DataFrame:
    """Load data from the IEA World Energy Balances.

    Parameters
    ----------
    provider : str
        First entry in :data:`.FILES`.
    edition : str
        Second entry in :data:`.FILES`.
    query_expr : str, optional
        Used with :meth:`pandas.DataFrame.query` to reduce the returned data.
    base_path : os.Pathlike, optional
        Path containing :data:`.FILES`. If not provided, locations within
        :mod:`message_data` or :mod:`message_ix_models` are used.

    Returns
    -------
    pandas.DataFrame
        The data frame has one column for each of :data:`.DIMS`, plus "Value".
    """
    files = FILES[(provider, edition)]

    # Identify a location that contains the `files`
    if path is None:
        if HAS_MESSAGE_DATA:
            path = private_data_path("iea")
        else:
            path = package_data_path("test", "iea")
            log.warning(f"Reading random data from {path}")

    assert path.joinpath(files[0]).exists()

    return iea_web_data_for_query(path, *files, query_expr=query_expr)


def generate_code_lists(
    provider: str, edition: str, output_path: Optional["os.PathLike"] = None
) -> None:
    """Extract structure from the data itself."""
    import sdmx.model.v21 as m

    from message_ix_models.util.sdmx import register_agency, write

    output_path = output_path or package_data_path("sdmx")

    IEA = m.Agency(
        id="IEA",
        name="International Energy Agency",
        contact=[m.Contact(uri=["https://iea.org"])],
    )
    register_agency(IEA)

    # Read the data
    if HAS_MESSAGE_DATA:
        path = private_data_path("iea")
    else:
        path = package_data_path("test", "iea")
        log.warning(f"Reading random data from {path}")
    data = iea_web_data_for_query(
        path, *FILES[(provider, edition)], query_expr="TIME > 0"
    )

    for concept_id in ("COUNTRY", "FLOW", "PRODUCT"):
        # Create a code list with the unique values from this dimension
        cl = m.Codelist(id=f"{concept_id}_{provider}", maintainer=IEA, version=edition)
        cl.extend(
            m.Code(id=code_id) for code_id in sorted(data[concept_id].dropna().unique())
        )
        write(cl, output_path)



"""Utilities for handling objects from :mod:`sdmx`."""

import logging
from datetime import datetime
from enum import Enum, Flag
from importlib.metadata import version
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Mapping, Optional, Union
from warnings import warn

import sdmx
import sdmx.message
from iam_units import registry
from sdmx.model.v21 import AnnotableArtefact, Annotation, Code, InternationalString

from .common import package_data_path

if TYPE_CHECKING:
    from os import PathLike

    import sdmx.model.common

log = logging.getLogger(__name__)

CodeLike = Union[str, Code]


# FIXME Reduce complexity from 13 → ≤11
def as_codes(  # noqa: C901
    data: Union[List[str], Dict[str, CodeLike]],
) -> List[Code]:
    """Convert `data` to a :class:`list` of |Code| objects.

    Various inputs are accepted:

    - :class:`list` of :class:`str`.
    - :class:`dict`, in which keys are :attr:`~sdmx.model.common.Code.id` and values are
      further :class:`dict` with keys matching other Code attributes.
    """
    # Assemble results as a dictionary
    result: Dict[str, Code] = {}

    if isinstance(data, list):
        # FIXME typing ignored temporarily for PR#9
        data = dict(zip(data, data))  # type: ignore [arg-type]
    elif not isinstance(data, Mapping):
        raise TypeError(data)

    for id, info in data.items():
        # Pass through Code; convert other types to dict()
        if isinstance(info, Code):
            result[info.id] = info
            continue
        elif isinstance(info, str):
            _info = dict(name=info)
        elif isinstance(info, Mapping):
            _info = dict(info)
        else:
            raise TypeError(info)

        # Create a Code object
        code = Code(
            id=str(id),
            name=_info.pop("name", str(id).title()),
        )

        # Store the description, if any
        try:
            code.description = InternationalString(value=_info.pop("description"))
        except KeyError:
            pass

        # Associate with a parent
        try:
            parent_id = _info.pop("parent")
        except KeyError:
            pass  # No parent
        else:
            result[parent_id].append_child(code)

        # Associate with any children
        for id in _info.pop("child", []):
            try:
                code.append_child(result[id])
            except KeyError:
                pass  # Not parsed yet

        # Convert other dictionary (key, value) pairs to annotations
        for id, value in _info.items():
            code.annotations.append(
                Annotation(id=id, text=value if isinstance(value, str) else repr(value))
            )

        result[code.id] = code

    return list(result.values())


def eval_anno(obj: AnnotableArtefact, id: str):
    """Retrieve the annotation `id` from `obj`, run :func:`eval` on its contents.

    .. deprecated:: 2023.9.12

       Use :meth:`sdmx.model.common.AnnotableArtefact.eval_annotation`, which provides
       the same functionality.
    """
    warn(
        "message_ix_models.util.eval_anno; use sdmx.model.common.AnnotableArtefact"
        ".eval_annotation() instead.",
        DeprecationWarning,
        stacklevel=2,
    )

    try:
        value = str(obj.get_annotation(id=id).text)
    except KeyError:  # No such attribute
        return None

    try:
        return eval(value, {"registry": registry})
    except Exception as e:  # Something that can't be eval()'d, e.g. a plain string
        log.debug(f"Could not eval({value!r}): {e}")
        return value


def make_enum(urn, base=Enum):
    """Create an :class:`.enum.Enum` (or `base`) with members from codelist `urn`."""
    # Read the code list
    cl = read(urn)

    # Ensure the 0 member is NONE, not any of the codes
    names = ["NONE"] if issubclass(base, Flag) else []
    names.extend(code.id for code in cl)

    # Create the class
    return base(urn, names)


def read(urn: str, base_dir: Optional["PathLike"] = None):
    """Read SDMX object from package data given its `urn`."""
    # Identify a path that matches `urn`
    base_dir = Path(base_dir or package_data_path("sdmx"))
    urn = urn.replace(":", "_")  # ":" invalid on Windows
    paths = sorted(
        set(base_dir.glob(f"*{urn}*.xml")) | set(base_dir.glob(f"*{urn.upper()}*.xml"))
    )

    if len(paths) > 1:
        log.info(
            f"Match {paths[0].relative_to(base_dir)} for {urn!r}; {len(paths) -1 } "
            "other result(s)"
        )

    try:
        with open(paths[0], "rb") as f:
            msg = sdmx.read_sdmx(f)
    except IndexError:
        raise FileNotFoundError(f"'*{urn}*.xml', '*{urn.upper()}*.xml' or similar")

    for _, cls in msg.iter_collections():
        try:
            return next(iter(msg.objects(cls).values()))
        except StopIteration:
            pass


def write(obj, base_dir: Optional["PathLike"] = None, basename: Optional[str] = None):
    """Store an SDMX object as package data."""
    base_dir = Path(base_dir or package_data_path("sdmx"))

    if isinstance(obj, sdmx.message.StructureMessage):
        msg = obj
        assert basename
    else:
        # Set the URN of the object
        obj.urn = sdmx.urn.make(obj)

        # Wrap the object in a StructureMessage
        msg = sdmx.message.StructureMessage()
        msg.add(obj)

        # Identify a path to write the file. ":" is invalid on Windows.
        basename = basename or obj.urn.split("=")[-1].replace(":", "_")

    msg.header = sdmx.message.Header(
        source=f"Generated by message_ix_models {version('message_ix_models')}",
        prepared=datetime.now(),
    )

    path = base_dir.joinpath(f"{basename}.xml")

    # Write
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_bytes(sdmx.to_xml(msg, pretty_print=True))

    log.info(f"Wrote {path}")


def register_agency(
    agency: "sdmx.model.common.Agency",
) -> "sdmx.model.common.AgencyScheme":
    """Add `agency` to the :class:`.AgencyScheme` "IIASA_ECE:AGENCIES"."""
    # Read the existing agency scheme
    as_ = read("IIASA_ECE:AGENCIES")

    if agency in as_:
        log.info(f"Replace or update existing {as_[agency.id]!r}")
        as_.items[agency.id] = agency
    else:
        as_.append(agency)

    log.info(f"Updated {as_!r}")

    # Write to file again
    write(as_)

    return as_



import logging
from copy import copy

from genno import Computer, Key

from message_ix_models.tools.exo_data import (
    ExoDataSource,
    iamc_like_data_for_query,
    register_source,
)
from message_ix_models.util import (
    HAS_MESSAGE_DATA,
    iter_keys,
    package_data_path,
    private_data_path,
)

__all__ = [
    "ADVANCE",
]

log = logging.getLogger(__name__)


#: Expected location of the ADVANCE WP2 data snapshot.
LOCATION = "advance", "advance_compare_20171018-134445.csv.zip"

#: Name of the data file within the archive.
NAME = "advance_compare_20171018-134445.csv"


@register_source
class ADVANCE(ExoDataSource):
    """Provider of exogenous data from the ADVANCE project database.

    To use data from this source, call :func:`.exo_data.prepare_computer` with the
    arguments:

    - `source`: "ADVANCE"
    - `source_kw` including:

      - "model": one of 12 codes including "MESSAGE".
      - "measure": one of 3080 codes for the "VARIABLE" dimension.
      - "scenario": one of 51 codes including "ADV3TRAr2_Base".
      - "name", optional: override :attr:`.ExoDataSource.name`.
      - "aggregate", optional: if :any:`True`, aggregate data from the ADVANCE native
        regions using ``n::groups`` (same behaviour as the base class). Otherwise, do
        not aggregate.

    Example
    -------
    >>> keys = prepare_computer(
    ...     context,
    ...     computer,
    ...     source="ADVANCE",
    ...     source_kw=dict(
    ...         measure="Transport|Service demand|Road|Freight",
    ...         model="MESSAGE",
    ...         scenario="ADV3TRAr2_Base",
    ...     ),
    ... )
    >>> result = computer.get(keys[0])

    Load the metadata packaged with :mod:`message_ix_models` to identify usable
    `source_kw`:

    >>> import sdmx
    >>> from message_ix_models.util import package_data_path
    >>>
    >>> msg = sdmx.read_sdmx(package_data_path("sdmx", "ADVANCE.xml"))
    >>> msg
    <sdmx.StructureMessage>
    <Header>
        prepared: '2023-11-03T21:51:46.052879'
        source: en: Generated by message_ix_models 2023.9.13.dev57+ga558c0b4.d20231011
        test: False
    Codelist (5): MODEL SCENARIO REGION VARIABLE UNIT
    >>> msg.codelist["MODEL].items
    {'AIM/CGE': <Code AIM/CGE>,
     'DNE21+': <Code DNE21+>,
     'GCAM': <Code GCAM>,
     'GEM-E3': <Code GEM-E3>,
     'IMACLIM V1.1': <Code IMACLIM V1.1>,
     'IMAGE': <Code IMAGE>,
     'MESSAGE': <Code MESSAGE>,
     'POLES ADVANCE': <Code POLES ADVANCE>,
     'REMIND': <Code REMIND>,
     'TIAM-UCL': <Code TIAM-UCL>,
     'WITCH': <Code WITCH>,
     'iPETS V.1.5': <Code iPETS V.1.5>}
    """

    id = "ADVANCE"

    def __init__(self, source, source_kw):
        if not source == self.id:
            raise ValueError(source)

        # Map the `measure` keyword to a string appearing in the data
        _kw = copy(source_kw)
        measure = _kw.pop("measure")
        self.variable = {
            "GDP": "GDP|PPP",
            "POP": "Population",
        }.get(measure, measure)

        # Store the model and scenario ID
        self.model = _kw.pop("model", None)
        self.scenario = _kw.pop("scenario", None)

        # Set the name of the returned quantity
        self.name = _kw.pop("name", "")

        self.aggregate = _kw.pop("aggregate", True)

        if len(_kw):
            raise ValueError(_kw)

    def __call__(self):
        # Assemble a query string
        query = " and ".join(
            [
                f"SCENARIO == {self.scenario!r}",
                f"VARIABLE == {self.variable!r}",
                f"MODEL == {self.model!r}" if self.model else "True",
            ]
        )
        log.debug(query)

        # Expected location of the ADVANCE WP2 data snapshot.
        parts = LOCATION
        if HAS_MESSAGE_DATA:
            path = private_data_path(*parts)
        else:
            path = package_data_path("test", *parts)
            log.warning(f"Reading random data from {path}")

        return iamc_like_data_for_query(
            path, query, archive_member=NAME, non_iso_3166="keep"
        )

    def transform(self, c: "Computer", base_key: Key) -> Key:
        """Prepare `c` to transform raw data from `base_key`.

        Unlike the base class version, this implementation only adds the aggregation
        step if :attr:`.aggregate` is :any:`True`.
        """
        k = iter_keys(base_key)

        k1 = base_key
        if self.aggregate:
            # Aggregate
            k1 = k()
            c.add(k1, "aggregate", base_key, "n::groups", keep=False)

        # Interpolate to the desired set of periods
        kw = dict(fill_value="extrapolate")
        k2 = k()
        c.add(k2, "interpolate", k1, "y::coords", kwargs=kw)

        return k2



import numpy as np
import pytest
from message_ix import make_df
from message_ix.models import MACRO

from message_ix_models import testing
from message_ix_models.model.emissions import add_tax_emission, get_emission_factors


def add_test_data(scenario):
    scenario.platform.add_unit("")
    with scenario.transact():
        scenario.add_set("node", "foo")
        scenario.add_set("type_emission", "TCE")

        name = "interestrate"
        df = make_df(name, year=scenario.set("year"), value=0.05, unit="")
        scenario.add_par(name, df)

        # Initialize drate
        MACRO.initialize(scenario)
        name = "drate"
        df = scenario.add_par(
            name, make_df(name, node=["World", "foo"], value=[0.05, 0.03], unit="")
        )


def test_add_tax_emission(request, caplog, test_context):
    test_context.regions = "R12"
    s = testing.bare_res(request, test_context, solved=False)
    add_test_data(s)

    value = 1.1

    add_tax_emission(s, value, drate_parameter="interestrate")

    # Retrieve the added data
    data = s.par("tax_emission").set_index("type_year")
    y0 = min(data.index)
    y_max = max(data.index)

    # First model period value is converted from [money] / t CO₂ to [money] / t C
    v0 = value * 44.0 / 12
    assert np.isclose(v0, data.loc[y0, "value"])

    # Final period value is the same, inflated by the number of intervening years
    assert np.isclose(v0 * 1.05 ** (int(y_max) - int(y0)), data.loc[y_max, "value"])

    # Same using drate
    add_tax_emission(s, value)
    # Warning is logged about multiple drates
    assert (
        "Using the first of multiple discount rates: drate=[0.05 0.03]"
        == caplog.messages[-1]
    )


@pytest.mark.parametrize(
    "units, exp_coal",
    (
        (None, 25.8),
        # Unit expressions and values appearing in the message_doc table
        ("tC / TJ", 25.8),
        ("t CO2 / TJ", 94.6),
        ("t C / kWa", 0.8142),
    ),
)
def test_get_emission_factors(units, exp_coal):
    # Data are loaded
    result = get_emission_factors(units=units)
    assert 8 == result.size

    # Expected values are obtained
    assert np.isclose(exp_coal, result.sel(c="coal").item(), rtol=1e-4)



import logging
import os
import platform
import re
from copy import deepcopy
from pathlib import Path

import click
import ixmp
import pytest
from message_ix import Scenario

from message_ix_models import Context


class TestContext:
    def test_get_instance(self, session_context):
        c = Context()
        assert c is Context.get_instance(-1)
        c.delete()

    def test_only(self):
        # Ensure at least 2 instances exist
        c2 = Context()

        with pytest.raises(IndexError, match=r"ambiguous: \d+ Context instances"):
            Context.only()

        c2.delete()

    def test_clone_to_dest(self, caplog, test_context):
        ctx = test_context

        platform_name = ctx.platform_info["name"]
        model_name = "foo model"
        scenario_name = "bar scenario"

        # Works with direct settings, no URL
        c = deepcopy(ctx)

        # Force the base scenario info to be empty
        c.scenario_info.clear()
        c.dest_scenario = dict(model=model_name, scenario=scenario_name)

        # Fails with create=False
        with pytest.raises(
            TypeError, match="missing 1 required positional argument: 'model'"
        ):
            c.clone_to_dest(create=False)

        # Succeeds with default create=True
        s = c.clone_to_dest()

        # Base scenario was created
        assert "Base scenario not given or found" in caplog.messages

        # Works with a URL to parse and no base scenario
        url = f"ixmp://{platform_name}/{model_name}/{scenario_name}"

        c2 = deepcopy(ctx)
        c2.scenario_info.clear()
        c2.dest = url
        s = c2.clone_to_dest()
        assert model_name == s.model and scenario_name == s.scenario

        del s

        # Works with a base scenario
        c2.handle_cli_args(url=url)
        c2.dest_scenario = dict(model="baz model", scenario="baz scenario")
        c2.dest_platform.clear()
        s = c2.clone_to_dest()

        assert s.model.startswith("baz") and s.scenario.startswith("baz")

    def test_dealias(self, caplog):
        """Aliasing works with :meth:`.Context.__init__`, :meth:`.Context.update`."""
        c = Context()
        c.update(regions="R99")
        assert [] == caplog.messages  # No log warnings for core Config, .model.Config
        assert "R99" == c.model.regions

        c = Context(regions="R98")
        assert [
            "Create a Config instance instead of passing ['regions'] to Context()"
        ] == caplog.messages
        caplog.clear()
        assert "R98" == c.model.regions == c.regions
        assert [] == caplog.messages  # No log warnings for access

    def test_default_value(self, test_context):
        # Setting is missing
        with pytest.raises(AttributeError):
            test_context.foo

        # setdefault() returns the new value
        assert test_context.setdefault("foo", 23) == 23

        # setdefault() returns the existing value
        assert test_context.setdefault("foo", 45) == 23

        # Attribute access works
        assert test_context.foo == 23

    def test_deepcopy(self, session_context):
        """Paths are preserved through deepcopy()."""
        ld = session_context.local_data

        c = deepcopy(session_context)

        assert ld == c.local_data

    def test_get_cache_path(self, pytestconfig, test_context):
        """cache_path() returns the expected output."""
        # One of two values depending on whether the user has given --local-cache
        assert (
            test_context.get_cache_path("pytest", "bar.pkl")
            in (
                test_context.local_data.joinpath("cache", "pytest", "bar.pkl"),
                Path(pytestconfig.cache.makedir("cache")).joinpath("pytest", "bar.pkl"),
            )
            or pytestconfig.option.local_cache
        )

    def test_get_local_path(self, tmp_path_factory, session_context):
        assert str(tmp_path_factory.mktemp("data").joinpath("foo", "bar")).replace(
            "data1", "data0"
        ) == str(session_context.get_local_path("foo", "bar"))

    def test_get_platform(self, session_context):
        assert isinstance(session_context.get_platform(), ixmp.Platform)
        assert isinstance(session_context.get_platform(reload=True), ixmp.Platform)

    def test_get_scenario(self, test_context):
        test_context.scenario_info = dict(model="model name", scenario="scenario name")
        with pytest.raises(ValueError):
            test_context.get_scenario()

    @pytest.mark.flaky(
        reruns=5,
        rerun_delay=2,
        condition="GITHUB_ACTIONS" in os.environ and platform.system() == "Darwin",
        reason="Flaky; see iiasa/message-ix-models#112",
    )
    def test_set_scenario(self, test_context):
        mp = test_context.get_platform()
        s = Scenario(mp, "foo", "bar", version="new")

        # set_scenario() updates Context.scenario_info
        test_context.scenario_info = dict()
        test_context.set_scenario(s)
        assert (
            dict(model="foo", scenario="bar", version=0) == test_context.scenario_info
        )

    def test_write_debug_archive(self, mix_models_cli):
        """:meth:`.write_debug_archive` works."""
        # Create a CLI command attached to the hidden "_test" group

        from message_ix_models.cli import cli_test_group
        from message_ix_models.util.click import temporary_command

        @click.command("write-debug-archive")
        @click.pass_obj
        def command(context):
            # Register one file to be archived
            p = context.core.local_data.joinpath("foo.txt")
            context.core.debug_paths.append(p)

            # Write some text to this file
            p.write_text("Here is some debug output in a file.")

            # Register a non-existent path
            context.core.debug_paths.append(p.with_name("bar.txt"))

            # Write the archive
            context.write_debug_archive()

        # Invoke the command; I/O occurs in a temporary directory
        with temporary_command(cli_test_group, command):
            result = mix_models_cli.invoke(["_test", "write-debug-archive"])

        # Output path is constructed as expected; file exists
        match = re.search(
            r"Write to: (.*main-_test-write-debug-archive-[\dabcdefT\-]+.zip)",
            result.output,
        )
        assert Path(match.group(1)).exists()

        # Log output is generated for the non-existent path in Context.debug_paths
        assert re.search(r"Not found: .*bar.txt", result.output)

    def test_handle_cli_args(self):
        p = "platform name"
        m = "model name"
        s = "scenario name"
        v = "42"

        args1 = dict(
            local_data=Path("foo", "bar"),
            platform=p,
            model_name=m,
            scenario_name=s,
            version=v,
        )

        expected = dict(
            local_data=args1["local_data"],
            platform_info=dict(name=p),
            scenario_info=dict(model=m, scenario=s, version=v),
        )

        ctx = Context()
        ctx.handle_cli_args(**args1)
        assert all(ctx[k] == v for k, v in expected.items())

        url = f"ixmp://{p}/{m}/{s}#{v}"
        args2 = args1.copy()
        args2["url"] = url

        with pytest.raises(click.BadOptionUsage, match="redundant with --url"):
            ctx.handle_cli_args(**args2)

        # New instance
        ctx.delete()
        ctx = Context()

        # Platform and scenario info are empty
        assert 0 == len(ctx["platform_info"]) == len(ctx["scenario_info"])

        ctx.handle_cli_args(url=url, local_data=args1["local_data"])

        # ixmp parse_url() converts the version number to an int
        expected["scenario_info"]["version"] = int(v)
        # url is also stored
        expected["url"] = url

        assert all(ctx[k] == v for k, v in expected.items()), ctx

        ctx.delete()

    def test_repr(self) -> None:
        c = Context()
        assert re.fullmatch("<Context object at [^ ]+ with 3 keys>", repr(c))

    def test_use_defaults(self, caplog):
        caplog.set_level(logging.INFO)

        c = Context()

        defaults = dict(foo=["foo2", "foo1", "foo3"], bar=["bar1", "bar3"])

        c.foo = "foo1"

        c.use_defaults(defaults)
        assert ["Use default bar=bar1"] == caplog.messages

        c.bar = "bar2"

        with pytest.raises(
            ValueError, match=re.escape("bar must be in ['bar1', 'bar3']; got bar2")
        ):
            c.use_defaults(defaults)

        c.delete()



"""Basic tests of the command line."""

import click

from message_ix_models.cli import cli_test_group
from message_ix_models.util.click import common_params, temporary_command


def test_default_path_cb(session_context, mix_models_cli):
    """Test :func:`.default_path_cb`."""

    # Create a hidden command and attach it to the CLI
    @click.command("default_path_cb")
    @common_params("rep_out_path")
    @click.pass_obj
    def func(ctx, rep_out_path):
        print(ctx["rep_out_path"])  # Print the value stored on the Context object

    # Command parameters: --local-data gives the local data path, but the --rep-out-path
    # option is *not* given
    cmd = [f"--local-data={session_context.local_data}", "_test", func.name]

    # …so default_path_cb() should supply "{local_data}/reporting_output".
    expected = session_context.local_data / "reporting_output"

    # Run the command
    with temporary_command(cli_test_group, func):
        result = mix_models_cli.assert_exit_0(cmd)

    # The value was stored on, and retrieved from, `ctx`
    assert result.output.startswith(f"{expected}\n")


def test_regions(mix_models_cli):
    """--regions=… used on both group and a command within the group.

    If the option is not provided to the inner command, the value given to the outer
    group should persist.
    """

    @click.group()
    @common_params("regions")
    def outer(regions):
        pass

    @outer.command()
    @common_params("regions")
    @click.pass_obj
    def inner(context, regions):
        print(context.model.regions)

    # Give the option for the outer group, but not for the inner command
    with temporary_command(cli_test_group, outer):
        result = mix_models_cli.assert_exit_0(
            ["_test", "outer", "--regions=ZMB", "inner"]
        )

    # Value given to the outer group is stored and available to the inner command
    assert "ZMB" == result.output.strip()


def test_store_context(mix_models_cli):
    """Test :func:`.store_context`."""

    # Create a hidden command and attach it to the CLI
    @click.command("store_context")
    @common_params("ssp")
    @click.pass_obj
    def func(ctx, ssp):
        print(ctx["ssp"])  # Print the value stored on the Context object

    # Run the command with a valid value
    with temporary_command(cli_test_group, func):
        result = mix_models_cli.assert_exit_0(["_test", func.name, "SSP2"])

    # The value was stored on, and retrieved from, `ctx`
    assert "SSP2\n" == result.output


def test_urls_from_file(mix_models_cli, tmp_path):
    """Test :func:`.urls_from_file` callback."""

    # Create a hidden command and attach it to the CLI
    @click.command("urls_from_file")
    @common_params("urls_from_file")
    @click.pass_obj
    def func(ctx, **kwargs):
        # Print the value stored on the Context object
        print("\n".join([s.url for s in ctx.core.scenarios]))

    # Create a temporary file with some scenario URLs
    text = """m/s#3
foo/bar#5
baz/qux#123
"""
    p = tmp_path.joinpath("scenarios.txt")
    p.write_text(text)

    # Run the command, referring to the temporary file
    with temporary_command(cli_test_group, func):
        result = mix_models_cli.assert_exit_0(
            ["_test", func.name, f"--urls-from-file={p}"]
        )

    # Scenario URLs are parsed to ScenarioInfo objects, and then can be reconstructed →
    # data is round-tripped
    assert text == result.output



"""Tools for World Bank data."""

import logging
from collections import defaultdict
from functools import lru_cache
from typing import TYPE_CHECKING, Dict, MutableMapping, Optional

import pandas as pd

if TYPE_CHECKING:
    import sdmx.model.common

log = logging.getLogger(__name__)


# FIXME Reduce complexity from 12 → ≤11
def assign_income_groups(  # noqa: C901
    cl_node: "sdmx.model.common.Codelist",
    cl_income_group: "sdmx.model.common.Codelist",
    method: str = "population",
    replace: Optional[Dict[str, str]] = None,
) -> None:
    """Annotate `cl_node` with income groups.

    Each node is assigned an |Annotation| with :py:`id="wb-income-group"`, according to
    the income groups of its children (countries), as reflected in `cl_income_group`
    (see :func:`.get_income_group_codelist`).

    Parameters
    ----------
    method : "population" or "count"
        Method for aggregation:

        - :py:`"population"` (default): the WB World Development Indicators (WDI) 2020
          population for each country is used as a weight, so that the node's income
          group is the income group of the plurality of the population of its children.
        - :py:`"count"`: each country is weighted equally, so that the node's income
          group is the mode (most frequently occurring value) of its childrens'.
    replace : dict
        Mapping from wb-income-group annotation text appearing in `cl_income_group` to
        texts to be attached to `cl_node`. Mapping two keys to the same value
        effectively combines or aggregates those groups. See :func:`.make_map`.

    Example
    -------
    Annotate the R12 node list with income group information, mapping high income
    countries (HIC) and upper-middle income countries (UMC) into one group and
    aggregating by population.

    >>> cl_node = get_codelist(f"node/R12")
    >>> cl_ig = get_income_group_codelist()
    >>> replace = make_map({"HIC": "HMIC", "UMC": "HMIC"})
    >>> assign_income_groups(cl_node, cl_ig, replace=replace)
    >>> cl_node["R12_NAM"].get_annotation(id="wb-income-group").text
    HMIC
    """

    import sdmx
    import sdmx.model.v21 as m

    replace = replace or dict()

    if method == "count":

        def get_weight(code: "sdmx.model.common.Code") -> float:
            """Weight of the country `code` in aggregation."""
            return 1.0

    elif method == "population":
        # Retrieve WB_WDI data for SERIES=SP_POP_TOTAL (Population, total)
        dm = sdmx.Client("WB_WDI").data(
            "WDI", key="A.SP_POP_TOTL.", params=dict(start_period=2020, end_period=2020)
        )

        # Convert to pd.Series with multi-index with levels: REF_AREA, SERIES, FREQ,
        # TIME_PERIOD. Because of the query, there is only 1 value for each unique
        # REF_AREA.
        df = sdmx.to_pandas(dm.data[0])

        def get_weight(code: "sdmx.model.common.Code") -> float:
            """Return a weight for the country `code`: its total population."""
            try:
                return df[code.id].item()
            except KeyError:
                # log.debug(f"No population data for {code!r}; omitted")
                return 0
    else:  # pragma: no cover
        raise ValueError(f"method={method!r}")

    weight_info = {}  # For debugging

    # Iterate over nodes
    for node in cl_node:
        if not len(node.child):
            continue  # Country → skip

        # Total weight of different income groups among `node`'s countries
        weight: MutableMapping[Optional[str], float] = defaultdict(lambda: 0.0)

        # Iterate over countries
        for country in node.child:
            # Identify the income group of `country` from an annotation
            try:
                ig = str(
                    cl_income_group[country.id]
                    .get_annotation(id="wb-income-group")
                    .text
                )
                # Apply replacement to `ig`
                ig = replace.get(ig, ig)
            except KeyError:
                # country.id is not in cl_income_group, or no such annotation
                ig = None

            weight[ig] += get_weight(country)

        if {None} == set(weight):
            continue  # World node → no direct children that are countries

        # Sort weights and group IDs from largest/first alphabetically to smallest/last
        weight_sorted = sorted([(-v, k) for k, v in weight.items()])
        weight_info[node.id] = pd.Series({k: -v for v, k in weight_sorted})

        # Identify the income group with the largest weight; not None
        _, ig = next(filter(lambda item: item[1] is not None, weight_sorted))

        try:
            # Remove any existing annotation
            node.pop_annotation(id="wb-income-group")
        except KeyError:
            pass

        # Annotate the node
        node.annotations.append(m.Annotation(id="wb-income-group", text=ig))

    log.debug(
        "(node, group) weights:\n"
        + pd.concat(weight_info, axis=1).fillna(0).to_string()
    )


def fetch_codelist(id: str) -> "sdmx.model.common.Codelist":
    """Retrieve code lists related to the WB World Development Indicators.

    In principle this could be done with :py:`sdmx.Client("WB_WDI").codelist(id)`, but
    the World Bank SDMX REST API does not support queries for a specific code list. See
    https://datahelpdesk.worldbank.org/knowledgebase/articles/1886701-sdmx-api-queries.

    :func:`fetch_codelist` retrieves http://api.worldbank.org/v2/sdmx/rest/codelist/WB/,
    the structure message containing *all* code lists; and extracts and returns the one
    with the given `id`.
    """
    import pooch
    import sdmx

    file = pooch.retrieve(
        url="http://api.worldbank.org/v2/sdmx/rest/codelist/WB/", known_hash=None
    )
    # Read the retrieved SDMX StructureMessage and extract the code list
    sm = sdmx.read_sdmx(file)

    return sm.codelist[id]


@lru_cache()
def get_income_group_codelist() -> "sdmx.model.common.Codelist":
    """Return a |Codelist| with World Bank income group information.

    The returned code list is a modified version of the one with URN
    ``…Codelist=WB:CL_REF_AREA_WDI(1.0)``, via :func:`.fetch_codelist`.

    This is augmented with information about the income group and lending category
    concepts as described at
    https://datahelpdesk.worldbank.org/knowledgebase/articles/906519

    The information is stored two ways:

    - Existing codes in the list like "HIC: High income" that designate groups of
      countries are associated with child codes that are designated as members of that
      country. These can be accessed at :attr:`Code.child
      <sdmx.model.common.Item.child>`.
    - Existing codes in the list like "ABW: Aruba" are annotated with:

      - :py:`id="wb-income-group"`: the URN of the income group code, for instance
        "urn:sdmx:org.sdmx.infomodel.codelist.Code=WB:CL_REF_AREA_WDI(1.0).HIC". This is
        an unambiguous reference to a code in the same list.
      - :py:`id="wb-lending-category"`: the name of the lending category, if any.

      These can be accessed using :attr:`Code.annotations
      <sdmx.model.common.AnnotableArtefact.annotations>`, :attr:`Code.get_annotation
      <sdmx.model.common.AnnotableArtefact.get_annotation>`, and other methods.
    """
    import pooch
    import sdmx.model.v21 as m

    cl = fetch_codelist("CL_REF_AREA_WDI")

    @lru_cache()
    def urn_for(name: str) -> str:
        """Return the URN of a code in `cl`, given its `name`."""
        for code in cl:
            if str(code.name) == name:
                return code.urn
        raise ValueError(name)  # pragma: no cover

    # Fetch the file containing the classification
    file = pooch.retrieve(
        url="https://datacatalogfiles.worldbank.org/ddh-published/0037712/DR0090755/"
        "CLASS.xlsx",
        known_hash="sha256:"
        "9b8452db52e391602c9e9e4d4ef4d254f505ce210ce6464497cf3e40002a3545",
    )

    # Open the retrieved file
    ef = pd.ExcelFile(file)

    # Read the "List of economies" sheet → store wb-{income-group,lending-category}
    tmp = (
        pd.read_excel(ef, sheet_name="List of economies")
        .drop(["Economy", "Region"], axis=1)
        .dropna(subset=["Income group"], axis=0)
        .set_index("Code")
    )
    for code in cl:
        try:
            row = tmp.loc[code.id, :]
        except KeyError:
            # log.debug(f"Not in 'List of economies' sheet: {code!r}")
            continue

        # Annotate wb-income-group; map a value like "Low income" to a URN
        code.annotations.append(
            m.Annotation(id="wb-income-group", text=urn_for(row["Income group"]))
        )

        try:
            code.annotations.append(
                m.Annotation(id="wb-lending-category", text=row["Lending category"])
            )
        except ValueError:
            pass  # text was None → no value

    # Read the "Groups" sheet → assign hierarchy
    for group_id, group_df in (
        pd.read_excel(ef, sheet_name="Groups")
        .drop(["GroupName", "CountryName", "Unnamed: 4"], axis=1)
        .groupby("GroupCode")
    ):
        try:
            # Identify the Code for this group ID
            group = cl[group_id]
        except KeyError:
            # log.debug(f"Group {group_id!r} is not in {cl}")
            continue

        for child_id in sorted(group_df["CountryCode"]):
            try:
                group.append_child(cl[child_id])
            except KeyError:
                # log.debug(f"No code for child {child_id!r}")
                continue

        # log.debug(f"{cl[group_id]}: {len(cl[group_id].child)} children")

    # Read "Notes" sheet → append to description of `cl`
    tmp = "\n\n".join(pd.read_excel(ef, sheet_name="Notes").dropna()["Notes"])

    # Ensure the "en" localization exists
    cl.description.localizations.setdefault("en", "")
    cl.description.localizations["en"] += (
        "\n\nThis code list has been modified from the official version by the "
        "'message-ix-models' Python package to add annotations and hierarchy parsed "
        "from the World Bank income groups and lending categories as described at "
        "https://datahelpdesk.worldbank.org/knowledgebase/articles/906519. The original"
        f" Excel file parsed includes the following descriptive text:\n\n{tmp}"
    )

    return cl


def make_map(
    source: Dict[str, str], expand_key_urn: bool = True, expand_value_urn: bool = False
) -> Dict[str, str]:
    """Prepare the :py:`replace` parameter of :func:`assign_income_groups`.

    The result has one (`key`, `value`) for each in `source`.

    Parameters
    ----------
    expand_key_urn : bool
        If :obj:`True` (the default), replace each `key` from `source` with the URN for
        the code in ``CL_REF_AREA_WDI`` with :py:`id=key`.
    expand_value_urn : bool
        If :obj:`True`, replace each `value` from `source` with the URN for the code in
        ``CL_REF_AREA_WDI`` with :py:`id=value`.
    """
    # Retrieve the code list
    cl = fetch_codelist("CL_REF_AREA_WDI")

    result = dict()
    for key, value in source.items():
        key = cl[key].urn if expand_key_urn else key
        value = cl[value].urn if expand_value_urn else value
        result[key] = value

    return result



import logging
from collections import defaultdict
from functools import lru_cache
from itertools import product
from typing import Optional, Tuple

import numpy as np
import pandas as pd
import xarray as xr
from sdmx.model.v21 import Code

from message_ix_models import Context
from message_ix_models.model.structure import get_codes
from message_ix_models.util import load_package_data

log = logging.getLogger(__name__)

# Configuration files
METADATA = [
    # Information about MESSAGE-water
    ("water", "config"),
    ("water", "set"),
    ("water", "technology"),
]


def read_config(context=None):
    """Read the water model configuration / metadata from file.

    Numerical values are converted to computation-ready data structures.

    Returns
    -------
    .Context
        The current Context, with the loaded configuration.
    """

    context = context or Context.get_instance(0)

    # if context.nexus_set == 'nexus':
    if "water set" in context:
        # Already loaded
        return context

    # Load water configuration
    for parts in METADATA:
        # Key for storing in the context
        key = " ".join(parts)

        # Actual filename parts; ends with YAML
        _parts = list(parts)
        _parts[-1] += ".yaml"

        context[key] = load_package_data(*_parts)

    return context


@lru_cache()
def map_add_on(rtype=Code):
    """Map addon & type_addon in ``sets.yaml``."""
    dims = ["add_on", "type_addon"]

    # Retrieve configuration
    context = read_config()

    # Assemble group information
    result = defaultdict(list)

    for indices in product(*[context["water set"][d]["add"] for d in dims]):
        # Create a new code by combining two
        result["code"].append(
            Code(
                id="".join(c.id for c in indices),
                name=", ".join(c.name for c in indices),
            )
        )

        # Tuple of the values along each dimension
        result["index"].append(tuple(c.id for c in indices))

    if rtype == "indexers":
        # Three tuples of members along each dimension
        indexers = zip(*result["index"])
        indexers = {
            d: xr.DataArray(list(i), dims="consumer_group")
            for d, i in zip(dims, indexers)
        }
        indexers["consumer_group"] = xr.DataArray(
            [c.id for c in result["code"]],
            dims="consumer_group",
        )
        return indexers
    elif rtype is Code:
        return sorted(result["code"], key=str)
    else:
        raise ValueError(rtype)


def add_commodity_and_level(df, default_level=None):
    # Add input commodity and level
    t_info = Context.get_instance()["water set"]["technology"]["add"]
    c_info = get_codes("commodity")

    @lru_cache()
    def t_cl(t):
        input = t_info[t_info.index(t)].anno["input"]
        # Commodity must be specified
        commodity = input["commodity"]
        # Use the default level for the commodity in the RES (per
        # commodity.yaml)
        level = (
            input.get("level", "water_supply")
            or c_info[c_info.index(commodity)].anno.get("level", None)
            or default_level
        )

        return commodity, level

    def func(row):
        row[["commodity", "level"]] = t_cl(row["technology"])
        return row

    return df.apply(func, axis=1)


def map_yv_ya_lt(
    periods: Tuple[int, ...], lt: int, ya: Optional[int] = None
) -> pd.DataFrame:
    """All meaningful combinations of (vintage year, active year) given `periods`.

    Parameters
    ----------
    labels : pandas.DataFrame
        Each column (dimension) corresponds to one in `df`. Each row represents one
        matched set of labels for those dimensions.
    lt : int, lifetime
    """
    if not ya:
        ya = periods[0]
        log.info(f"First active year set as {ya!r}")
    if not lt:
        raise ValueError("Add a fixed lifetime parameter 'lt'")

    # The following lines are the same as
    # message_ix.tests.test_feature_vintage_and_active_years._generate_yv_ya

    # - Create a mesh grid using numpy built-ins
    # - Take the upper-triangular portion (setting the rest to 0)
    # - Reshape
    data = np.triu(np.meshgrid(periods, periods, indexing="ij")).reshape((2, -1))
    # Filter only non-zero pairs
    df = pd.DataFrame(
        filter(sum, zip(data[0, :], data[1, :])),
        columns=["year_vtg", "year_act"],
        dtype=np.int64,
    )

    # Select values using the `ya` and `lt` parameters
    return df[(ya <= df.year_act) & (df.year_act - df.year_vtg <= lt)]



import logging
import re

import pytest

from message_ix_models.util._logging import mark_time, silence_log


def test_mark_time(caplog):
    # Call 3 times
    mark_time()  # Will only log if already called during the course of another test
    mark_time()
    mark_time()

    # Either 2 or 3 records
    assert len(caplog.records) in (2, 3)

    # Each message matches the expected format
    assert all(re.match(r" \+\d+\.\d = \d+\.\d seconds", m) for m in caplog.messages)


class TestQueueListener:
    #: Number of log messages to emit.
    N = 1_000

    #: Number of times to run the test.
    k = 4

    @pytest.mark.parametrize(
        "method",
        ("click", pytest.param("subprocess", marks=pytest.mark.skip(reason="Slow."))),
    )
    @pytest.mark.parametrize("k", range(k))
    def test_flush(self, caplog, mix_models_cli, method, k):
        """Test logging in multiple processes, multiple threads, and with :mod:`click`.

        With pytest-xdist, these :attr:`k` test cases will run in multiple processes.
        Each process will have its main thread, and the thread of the QueueListener.
        The test ensures that all :attr:`N` log records emitted by the :py:`func()` are
        "flushed" from the queue, transferred to stdout by the :class:`.StreamHandler`
        and captured by the :class:`.CliRunner`.
        """

        # Run the command, capture output
        # See message_ix_models.cli._log_threads
        result = mix_models_cli.assert_exit_0(
            ["_test", "log-threads", str(k), str(self.N)], method=method
        )

        # All records are emitted; the last record ends with N - 1
        assert result.output.rstrip().endswith(f"{self.N - 1}"), result.output.split(
            "\n"
        )[-2:]


def test_silence_log(caplog):
    # An example logger
    log = logging.getLogger("message_ix_models.model")

    msg = "Here's a warning!"

    # pytest caplog fixture picks up warning messages
    log.warning(msg)
    assert [msg] == caplog.messages

    caplog.clear()

    # silence_log() hides the messages
    with silence_log():
        log.warning(msg)

    assert [
        "Set level=40 for logger(s): message_data message_ix_models",
        "…restored.",
    ] == caplog.messages
    caplog.clear()

    # After the "with" block, logging is restored
    log.warning(msg)
    assert [msg] == caplog.messages



import logging
from collections import defaultdict
from copy import copy
from functools import partial
from itertools import product
from typing import List, Mapping, MutableMapping, Sequence

import message_ix
import pandas as pd
from sdmx.model.common import Annotation, Code

from message_ix_models import ScenarioInfo, Spec
from message_ix_models.model.build import apply_spec
from message_ix_models.util import (
    broadcast,
    make_io,
    make_matched_dfs,
    make_source_tech,
    merge_data,
    nodes_ex_world,
    same_node,
)

log = logging.getLogger(__name__)


def add(
    scenario: message_ix.Scenario,
    groups: Sequence[Code],
    technologies: Sequence[Code],
    template: Code,
    **options,
) -> Spec:
    """Add disutility formulation to `scenario`."""
    # Generate the spec given the configuration options
    spec = get_spec(groups, technologies, template)

    # Apply spec and add data
    apply_spec(scenario, spec, partial(get_data, spec=spec), **options)

    return spec


def get_spec(
    groups: Sequence[Code], technologies: Sequence[Code], template: Code
) -> Spec:
    """Get a spec for a disutility formulation.

    Parameters
    ----------
    groups : list of |Code|
        Identities of the consumer groups with distinct disutilities.
    technologies : list of |Code|
        The technologies to which the disutilities are applied.
    template : |Code|

    """
    s = Spec()

    s.require.set["technology"].extend(technologies)

    # Disutility commodity and source
    s.add.set["commodity"] = [Code(id="disutility")]
    s.add.set["technology"] = [Code(id="disutility source")]

    # Disutility is unitless
    # NB this value is currently ignored by .build.apply_spec(). See #45.
    s.add.set["unit"].append("")

    # Unrelated annotations in the template
    other_anno = list(
        filter(lambda a: a.id not in ("input", "output"), template.annotations)
    )

    # Add conversion technologies
    for t, g in product(technologies, groups):
        # String formatting arguments
        fmt = dict(technology=t, group=g)

        # Format each field in the "input" and "output" annotations
        input = {
            k: v.format(**fmt) for k, v in template.eval_annotation(id="input").items()
        }
        output = {
            k: v.format(**fmt) for k, v in template.eval_annotation(id="output").items()
        }

        # - Format the ID string from the template
        # - Create new "input" and "output" annotations
        # - Copy other annotations unmodified
        t_code = Code(
            id=template.id.format(**fmt),
            annotations=[
                Annotation(id="input", text=repr(input)),
                Annotation(id="output", text=repr(output)),
            ]
            + [copy(a) for a in other_anno],
        )

        # "commodity" set elements to add
        s.add.set["commodity"].extend([input["commodity"], output["commodity"]])

        # "technology" set elements to add
        t_code.annotations.append(Annotation(id="input", text=repr(input)))
        s.add.set["technology"].append(t_code)

    # Deduplicate "commodity" set elements
    s.add.set["commodity"] = sorted(map(str, set(s.add.set["commodity"])))

    return s


def get_data(scenario, spec, **kwargs) -> Mapping[str, pd.DataFrame]:
    """Get data for disutility formulation.

    Calls :meth:`data_conversion` and :meth:`data_source`.

    Parameters
    ----------
    spec : dict
        The output of :meth:`get_spec`.
    """
    if len(kwargs):
        log.warning(f"Ignore {repr(kwargs)}")

    info = ScenarioInfo(scenario)

    # Get conversion technology data
    data = data_conversion(info, spec)

    # Get and append source data
    merge_data(data, data_source(info, spec))

    return data


def dp_for(col_name: str, info: ScenarioInfo) -> pd.Series:  # pragma: no cover
    """:meth:`pandas.DataFrame.assign` helper for ``duration_period``.

    Returns a callable to be passed to :meth:`pandas.DataFrame.assign`. The callable
    takes a data frame as the first argument, and returns a :class:`pandas.Series`
    based on the ``duration_period`` parameter in `info`, aligned to `col_name` in the
    data frame.

    Currently (2021-04-07) unused.
    """

    def func(df):
        return df.merge(info.par["duration_period"], left_on=col_name, right_on="year")[
            "value_y"
        ]

    return func


def data_conversion(info, spec: Spec) -> MutableMapping[str, pd.DataFrame]:
    """Generate input and output data for disutility conversion technologies."""
    common = dict(
        mode="all",
        year_vtg=info.Y,
        year_act=info.Y,
        # No subannual detail
        time="year",
        time_origin="year",
        time_dest="year",
    )

    # Use the spec to retrieve information
    technology: List[Code] = spec.add.set["technology"]

    # Data to return
    data0: Mapping[str, List[pd.DataFrame]] = defaultdict(list)

    # Loop over conversion technologies
    for t in technology:
        # Use the annotations on the technology Code to get information about the
        # commodity, level, and unit
        input = t.eval_annotation(id="input")
        output = t.eval_annotation(id="output")
        if None in (input, output):
            if t.id == "disutility source":
                continue  # Data for this tech is from data_source()
            else:  # pragma: no cover
                raise ValueError(t)  # Error in user input

        # Make input and output data frames
        i_o = make_io(
            (input["commodity"], input["level"], input["unit"]),
            (output["commodity"], output["level"], output["unit"]),
            1.0,
            on="output",
            technology=t.id,
            **common,
        )
        for par, df in i_o.items():
            if par == "input":
                # Add input of disutility
                df = pd.concat(
                    [df, df.assign(commodity="disutility", unit="-")], ignore_index=True
                )

            data0[par].append(df)

    # - Concatenate to a single data frame per parameter
    # - Broadcast across nodes
    data = {
        par: pd.concat(dfs, ignore_index=True)
        .pipe(broadcast, node_loc=nodes_ex_world(info.N))
        .pipe(same_node)
        for par, dfs in data0.items()
    }

    # Create data for capacity_factor
    data.update(make_matched_dfs(base=data["input"], capacity_factor=1.0))

    return data


def data_source(info, spec) -> Mapping[str, pd.DataFrame]:
    """Generate data for a technology that emits the “disutility” commodity."""
    # List of input levels where disutility commodity must exist
    levels = set()
    for t in spec["add"].set["technology"]:
        input = t.eval_annotation(id="input")
        if input:
            levels.add(input["level"])

    log.info(f"Generate disutility on level(s): {repr(levels)}")

    # Use default capacity_factor = 1.0
    result = make_source_tech(
        info,
        common=dict(
            commodity="disutility",
            mode="all",
            technology="disutility source",
            time="year",
            time_dest="year",
            unit="-",
        ),
        output=1.0,
        var_cost=1.0,
    )
    result["output"] = result["output"].pipe(broadcast, level=sorted(levels))

    return result



import re
from collections import defaultdict
from copy import deepcopy
from pathlib import Path

import pytest
from iam_units import registry
from sdmx.model.v21 import Annotation, Code

from message_ix_models.model.structure import (
    codelists,
    generate_set_elements,
    get_codes,
    get_region_codes,
    process_commodity_codes,
    process_units_anno,
)
from message_ix_models.util import as_codes


@pytest.mark.parametrize(
    "kind, exp",
    [
        (
            "node",
            {
                "ADVANCE",
                "B210-R11",
                "ISR",
                "R11",
                "R12",
                "R14",
                "R17",
                "R20",
                "R32",
                "RCP",
                "ZMB",
            },
        ),
        ("relation", {"A", "B", "CD-LINKS"}),
        ("year", {"A", "B"}),
    ],
)
def test_codelists(kind, exp):
    """:func:`codelists` returns the expected IDs."""
    assert exp == set(codelists(kind))


class TestGetCodes:
    """Test :func:`get_codes` for different code lists."""

    @pytest.mark.parametrize(
        "name",
        (
            "cd_links/unit",
            "commodity",
            "level",
            "node/ISR",
            "node/R11",
            "node/R14",
            "node/R17",
            "node/R20",
            "node/R32",
            "node/RCP",
            "node/ZMB",
            "technology",
            "relation/A",
            "relation/B",
            "relation/CD-LINKS",
            "year/A",
            "year/B",
        ),
    )
    def test_get_codes(self, name):
        """The included code lists can be loaded."""
        get_codes(name)

    def test_hierarchy(self):
        """get_codes() returns objects with the expected hierarchical relationship."""
        codes = get_codes("node/R11")

        AUT = codes[codes.index("AUT")]
        R11_WEU = codes[codes.index("R11_WEU")]
        World = codes[codes.index("World")]

        assert R11_WEU is AUT.parent
        assert AUT in R11_WEU.child

        assert World is R11_WEU.parent
        assert R11_WEU in World.child

    def test_commodities(self):
        data = get_codes("commodity")

        # Some expected commodities are present
        for check in "coal", "electr":
            assert check in data

        # Units for one commodity can be retrieved and parsed
        g = dict(registry=registry)
        coal = data[data.index("coal")]
        assert isinstance(coal.eval_annotation("units", globals=g), registry.Unit)

        # Descriptions are parsed without new lines
        crudeoil = data[data.index("crudeoil")]
        assert "\n" not in str(crudeoil.description)

        # Processing a second time does not double-wrap the unit expressions
        process_commodity_codes(data)
        coal = data[data.index("coal")]
        assert isinstance(coal.eval_annotation("units", globals=g), registry.Unit)

    def test_levels(self):
        data = get_codes("level")

        # Some expected commodities are present
        for check in "primary", "useful":
            assert check in data

        # Descriptions are parsed without new lines
        assert "\n" not in str(data[data.index("primary")].description)

    @pytest.mark.parametrize(
        "codelist, to_check, length, member",
        [
            ("RCP", "R5_MAF", 69, "CIV"),
            ("R11", "R11_AFR", 50, "CIV"),
            ("R17", "R17_AFR", 50, "CIV"),
            ("R17", "R17_CHN", 2, "CHN"),
            ("R20", "R20_AFR", 50, "CIV"),
            ("R20", "R20_CHN", 2, "CHN"),
            ("R14", "R14_AFR", 51, "CIV"),
            ("R32", "R32SSA-L", 41, "CIV"),
        ],
    )
    def test_nodes(self, codelist, to_check, length, member):
        """Tests of node codelists."""
        # Node codelist can be loaded
        data = get_codes(f"node/{codelist}")

        # List contains a particular region
        assert to_check in data

        # Region contains the correct number of countries
        code = data[data.index(to_check)]
        assert len(code.child) == length

        # A specific country is present in the region
        assert member in code.child

        # get_region_codes() also works
        region_codes = get_region_codes(codelist)
        assert to_check in region_codes
        # Does not include codes that are not children of "World"
        assert member not in region_codes

    def test_node_historic_country(self):
        """get_codes() handles ISO 3166 alpha-3 codes for historic countries."""
        assert "SCG" in get_codes("node/R11")

    @pytest.mark.parametrize(
        "codelist, length, expected",
        [
            ("A", 20, {"emission_factor": 15, "renewable formulation": 2}),
            ("B", 319, {}),
            ("CD-LINKS", 314, {"removed": 80}),
        ],
    )
    def test_relation(self, codelist, length, expected):
        # Codelist can be loaded
        data = get_codes(f"relation/{codelist}")

        # Number of codes is as expected
        assert length == len(data)

        # Count codes by "group" annotation
        group_count = defaultdict(lambda: 0)
        for code in data:
            try:
                group = str(code.get_annotation(id="group").text)
            except KeyError:
                continue
            group_count[group] += 1

        # print(groups)

        # Each group has the expected number of members
        for group, N in expected.items():
            assert N == group_count[group]

    def test_technologies(self):
        # Retrieve the tech info without calling technologies.cli
        data = get_codes("technology")

        # Check the length of the returned dataframe
        assert len(data) == 377

        # Get info on a certain technology
        h2_fc_trp = data[data.index("h2_fc_trp")]
        assert ["transport", "useful"] == eval(
            str(h2_fc_trp.get_annotation(id="output").text)
        )

        # Check that the default value for 'vintaged' is False when omitted from the
        # YAML file
        elec_exp = data[data.index("elec_exp")]
        assert False is eval(str(elec_exp.get_annotation(id="vintaged").text))

    @pytest.mark.parametrize("codelist, length", [("A", 16), ("B", 28)])
    def test_year(self, codelist, length):
        """Year code lists can be loaded and contain the correct number of codes.

        :seealso: :meth:`.TestScenarioInfo.test_year_from_codes`.
        """
        # Year codelist can be loaded
        data = get_codes(f"year/{codelist}")

        # List contains the expected number of codes
        assert len(data) == length


def test_cli_techs(session_context, mix_models_cli):
    """Test the `techs` CLI command."""
    # Command runs without error
    result = mix_models_cli.assert_exit_0(["techs"])

    # Result test
    assert "\n[5 rows x 8 columns]\n" in result.output

    # Path to the temporary file written by the command
    path = Path(re.search("Write to (.*.csv)", result.output)[1])

    # File was written in the local data directory
    assert Path("technology.csv") == path.relative_to(session_context.local_data)

    # File was written with the expected contents
    assert path.read_text().startswith(
        "id,name,description,type,vintaged,sector,output,input\n"
        "CF4_TCE,CF4_TCE,Tetrafluoromethane (CF4) Total Carbon Emissions,"
        "primary,False,dummy,\"['dummy', 'primary']\",\n"
    )


GSE_DATA = {
    "colour": {"add": as_codes(["blue", "green", "red"])},
    # This is configuration to generate codes based on the colour set
    "technology": {"add": [Code(id="foo-{colour.id}"), Code(id="bar")]},
}


@pytest.mark.parametrize(
    "colour_expr, expected",
    [
        (None, {"foo-blue", "foo-green", "foo-red", "bar"}),
        ("red", {"bar"}),  # "red" has no .child codes
        ("^.*(?<!red)$", {"foo-blue", "foo-green", "bar"}),
    ],
)
def test_generate_set_elements(colour_expr, expected):
    data = deepcopy(GSE_DATA)

    # Add a _generate annotation
    data["technology"]["add"][0].annotations = [
        Annotation(id="_generate", text=repr(dict(colour=colour_expr)))
    ]

    # Code runs
    generate_set_elements(data, "technology")

    # Codes are generated according to the contents of the _generate annotation
    assert expected == set(map(str, data["technology"]["add"]))


def test_process_units_anno():
    # Prepare 2 codes: the parent has a units annotation, the child has none
    codes = as_codes({"foo": {"units": "kg"}, "bar": {"parent": "foo"}})

    process_units_anno("", codes[0])

    # Parents' units are propagated to the child
    assert registry.Unit("kg") == codes[1].eval_annotation(
        "units", dict(registry=registry)
    )



def test_create_bare(mix_models_cli):
    """The ``res create-bare`` CLI command can be invoked."""
    # "--nodes" is not a required option, but we give it anyway to test the CLI code
    # that handles it
    mix_models_cli.assert_exit_0(["res", "create-bare", "--nodes=R11"])



import logging
from collections import ChainMap, defaultdict
from datetime import datetime
from functools import partial, update_wrapper
from importlib.metadata import version
from itertools import count
from typing import (
    TYPE_CHECKING,
    Callable,
    Collection,
    Dict,
    Mapping,
    MutableMapping,
    Optional,
    Protocol,
    Sequence,
    Union,
)

import message_ix
import pandas as pd
import pint

from ._convert_units import convert_units, series_of_pint_quantity
from ._logging import mark_time, preserve_log_level, silence_log
from .cache import cached
from .common import (
    HAS_MESSAGE_DATA,
    MESSAGE_DATA_PATH,
    MESSAGE_MODELS_PATH,
    Adapter,
    MappingAdapter,
    load_package_data,
    load_private_data,
    local_data_path,
    package_data_path,
    private_data_path,
)
from .node import adapt_R11_R12, adapt_R11_R14, identify_nodes, nodes_ex_world
from .scenarioinfo import ScenarioInfo, Spec
from .sdmx import CodeLike, as_codes, eval_anno

if TYPE_CHECKING:
    import genno

__all__ = [
    "HAS_MESSAGE_DATA",
    "MESSAGE_DATA_PATH",
    "MESSAGE_MODELS_PATH",
    "Adapter",
    "MappingAdapter",
    "adapt_R11_R12",
    "adapt_R11_R14",
    "add_par_data",
    "aggregate_codes",
    "as_codes",
    "broadcast",
    "cached",
    "check_support",
    "convert_units",
    "copy_column",
    "datetime_now_with_tz",
    "eval_anno",
    "ffill",
    "identify_nodes",
    "iter_keys",
    "load_package_data",
    "load_private_data",
    "local_data_path",
    "make_io",
    "make_matched_dfs",
    "make_source_tech",
    "mark_time",
    "maybe_query",
    "merge_data",
    "minimum_version",
    "package_data_path",
    "preserve_log_level",
    "private_data_path",
    "replace_par_data",
    "same_node",
    "same_time",
    "series_of_pint_quantity",
    "show_versions",
    "silence_log",
    "strip_par_data",
]

log = logging.getLogger(__name__)


def add_par_data(
    scenario: message_ix.Scenario,
    data: Mapping[str, pd.DataFrame],
    dry_run: bool = False,
):
    """Add `data` to `scenario`.

    Parameters
    ----------
    data
        Dict with keys that are parameter names, and values are pd.DataFrame or other
        arguments
    dry_run : bool, *optional*
        Only show what would be done.

    See also
    --------
    strip_par_data
    """
    # TODO optionally add units automatically
    # TODO allow units column entries to be pint.Unit objects

    total = 0

    for par_name, values in data.items():
        N = values.shape[0]
        log.info(f"{N} rows in {repr(par_name)}")
        log.debug("\n" + values.to_string(max_rows=5))

        total += N

        if dry_run:
            continue

        # Work around iiasa/ixmp#425
        values["unit"] = values["unit"].str.replace("^$", "-", regex=True)

        try:
            scenario.add_par(par_name, values)
        except Exception:  # pragma: no cover
            print(values.head())
            raise

    return total


def aggregate_codes(df: pd.DataFrame, dim: str, codes):  # pragma: no cover
    """Aggregate `df` along dimension `dim` according to `codes`."""
    raise NotImplementedError

    # Construct an inverse mapping
    mapping = {}
    for code in codes:
        mapping.update({child.id: code.id for child in code.child})

    for key, group_series in df.groupby(dim):
        print(key, group_series.replace({dim: mapping}))


def broadcast(
    df: pd.DataFrame, labels: Optional[pd.DataFrame] = None, **kwargs
) -> pd.DataFrame:
    """Fill missing data in `df` by broadcasting.

    :func:`broadcast` is suitable for use with partly-filled data frames returned by
    :func:`.message_ix.util.make_df`, with 1 column per dimension, plus a "value"
    column. It is also usable with :meth:`pandas.DataFrame.pipe` for chained operations.

    `labels` (if any) are handled first: one copy or duplicate of `df` is produced for
    each row (set of labels) in this argument. Then, `kwargs` are handled;
    :func:`broadcast` returns one copy for each element in the cartesian product of the
    dimension labels given by `kwargs`.

    Parameters
    ----------
    labels : pandas.DataFrame
        Each column (dimension) corresponds to one in `df`. Each row represents one
        matched set of labels for those dimensions.
    kwargs
        Keys are dimensions. Values are labels along that dimension to fill.

    Returns
    -------
    pandas.DataFrame
        The length is either 1 or an integer multiple of the length of `df`.

    Raises
    ------
    ValueError
        if any of the columns in `labels` or `kwargs` are not present in `df`, or if
        those columns are present but not empty.

    Examples
    --------
    >>> from message_ix import make_df
    >>> from message_ix_models.util import broadcast
    # Create a base data frame with some empty columns
    >>> base = make_df("input", technology="t", value=[1.1, 2.2])
    # Broadcast (duplicate) the data across 2 dimensions
    >>> df = base.pipe(broadcast, node_loc=["node A", "node B"], mode=["m0", "m1"])
    # Show part of the result
    >>> df.dropna(axis=1)
      mode node_loc technology  value
    0   m0   node A          t    1.1
    1   m0   node A          t    2.2
    2   m0   node B          t    1.1
    3   m0   node B          t    2.2
    4   m1   node A          t    1.1
    5   m1   node A          t    2.2
    6   m1   node B          t    1.1
    7   m1   node B          t    2.2
    """

    def _check_dim(d):
        try:
            if not df[d].isna().all():
                raise ValueError(f"Dimension {d} was not empty\n\n{df.head()}")
        except KeyError:
            raise ValueError(f"Dimension {d} not among {list(df.columns)}")

    # Broadcast using matched labels for 1+ dimensions from a data frame
    if labels is not None:
        # Check the dimensions
        for dim in labels.columns:
            _check_dim(dim)
        # Concatenate 1 copy of `df` for each row in `labels`
        df = pd.concat(
            [df.assign(**row) for _, row in labels.iterrows()],
            ignore_index=True,
            sort=False,
        )

    # Next, broadcast other dimensions given as keyword arguments
    for dim, levels in kwargs.items():
        _check_dim(dim)
        if len(levels) == 0:
            log.debug(
                f"Don't broadcast over {repr(dim)}; labels {levels} have length 0"
            )
            continue

        # - Duplicate the data
        # - Drop the existing column named 'dim'
        # - Re-add the column from the constructed MultiIndex
        # - Reindex for sequential row numbers
        df = (
            pd.concat([df] * len(levels), keys=levels, names=[dim], sort=False)
            .drop(dim, axis=1)
            .reset_index(dim)
            .reset_index(drop=True)
        )
    return df


def check_support(context, settings=dict(), desc: str = "") -> None:
    """Check whether a Context is compatible with certain `settings`.

    Raises
    ------
    :class:`NotImplementedError`
        if any `context` value for a key of `settings` is not among the values in
        `settings`.
    :class:`KeyError`
        if the key is not set on `context` at all.

    See also
    --------
    :ref:`check-support`

    """
    __tracebackhide__ = True
    for key, values in settings.items():
        if context[key] not in values:
            raise NotImplementedError(
                f"{desc} for {repr(values)}; got {repr(context[key])}"
            )


def copy_column(column_name):
    """For use with :meth:`pandas.DataFrame.assign`.

    Examples
    --------
    Modify `df` by filling the column 'baz' with the value ``3``, and copying the column
    'bar' into column 'foo'.

    >>> df.assign(foo=copy_column('bar'), baz=3)

    Note that a similar assignment can be achieved with :meth:`~pandas.DataFrame.eval`:

    >>> df.eval("foo = bar")

    :func:`copy_column` is useful in the context of more complicated calls to
    :meth:`~pandas.DataFrame.assign`.
    """
    return lambda df: df[column_name]


def datetime_now_with_tz() -> datetime:
    """Current date and time with time zone information."""
    tz = datetime.now().astimezone().tzinfo
    return datetime.now(tz)


def ffill(
    df: pd.DataFrame, dim: str, values: Sequence[CodeLike], expr: Optional[str] = None
) -> pd.DataFrame:
    """Forward-fill `df` on `dim` to cover `values`.

    Parameters
    ----------
    df : pandas.DataFrame
        Data to fill forwards.
    dim : str
        Dimension to fill along. Must be a column in `df`.
    values : list of str
        Labels along `dim` that must be present in the returned data frame.
    expr : str, *optional*
        If provided, :meth:`.DataFrame.eval` is called. This can be used to assign one
        column to another. For instance, if `dim` == "year_vtg" and `expr` is "year_act
        = year_vtg", then forward filling is performed along the "year_vtg" dimension/
        column, and then the filled values are copied to the "year_act" column.
    """
    if dim in ("value", "unit"):
        raise ValueError(dim)

    # Mapping from (values existing in `df`) -> equal or greater members of `values`
    mapping = defaultdict(set)
    last_seen = None
    for v in sorted(set(values) | set(df[dim].unique())):
        if v in df[dim].unique():
            last_seen = v
        mapping[last_seen].add(v)

    def _maybe_eval(df):
        return df.eval(expr) if expr is not None else df

    dfs = [df]
    for key, group_df in df.groupby(dim):
        for new_label in sorted(mapping[key])[1:]:
            # Duplicate the data; assign the new_label to `dim`
            dfs.append(group_df.assign(**{dim: new_label}).pipe(_maybe_eval))

    return pd.concat(dfs, ignore_index=True)


class KeyIterator(Protocol):
    def __call__(self) -> "genno.Key": ...


def iter_keys(base: "genno.Key") -> KeyIterator:
    """Return an iterator over a sequence of keys starting with `base_key`.

    This can be used for shorthand when constructing sequences of :mod:`genno`
    computations.

    Example
    -------
    >>> base_key = genno.Key("foo:a-b-c")
    >>> k = iter_keys(base_key)
    >>> k()
    <foo:a-b-c:0>
    >>> k()
    <foo:a-b-c:1>
    >>> k()
    <foo:a-b-c:2>
    """
    return partial(next, map(lambda i: base + str(i), count()))


def iter_parameters(set_name, scenario: Optional["message_ix.Scenario"] = None):
    """Iterate over MESSAGEix parameters with *set_name* as a dimension.

    .. deprecated:: 2023.11
       Use :meth:`ixmp.Scenario.par_list` with the :py:`indexed_by=...` argument
       instead.
    """
    try:
        assert scenario is not None
        yield from scenario.items(indexed_by=set_name, par_data=False)
    except (TypeError, AssertionError):  # ixmp < 3.8.0
        import message_ix.models

        for name, info in message_ix.models.MESSAGE_ITEMS.items():
            if info["ix_type"] == "par" and set_name in info["idx_sets"]:
                yield name


def make_io(src, dest, efficiency, on="input", **kwargs):
    """Return input and output data frames for a 1-to-1 technology.

    Parameters
    ----------
    src : tuple (str, str, str)
        Input commodity, level, unit.
    dest : tuple (str, str, str)
        Output commodity, level, unit.
    efficiency : float
        Conversion efficiency.
    on : 'input' or 'output'
        If 'input', `efficiency` applies to the input, and the output, thus the activity
        level of the technology, is in dest[2] units. If 'output', the opposite.
    kwargs
        Passed to :func:`~message_ix.make_df`.

    Returns
    -------
    dict (str -> pd.DataFrame)
        Keys are 'input' and 'output'; values are data frames.
    """
    return dict(
        input=message_ix.make_df(
            "input",
            commodity=src[0],
            level=src[1],
            unit=src[2],
            value=efficiency if on == "input" else 1.0,
            **kwargs,
        ),
        output=message_ix.make_df(
            "output",
            commodity=dest[0],
            level=dest[1],
            unit=dest[2],
            value=1.0 if on == "input" else efficiency,
            **kwargs,
        ),
    )


def make_matched_dfs(
    base: MutableMapping, **par_value: Union[float, pint.Quantity]
) -> Dict[str, pd.DataFrame]:
    """Return data frames derived from `base` for multiple parameters.

    Creates one data frame per keyword argument.

    Parameters
    ----------
    base : pandas.DataFrame, dict, etc.
        Used to populate other columns of each data frame. Duplicates—which occur when
        the target parameter has fewer dimensions than `base`—are dropped.
    par_values :
        Argument names (e.g. ‘fix_cost’) are passed to :func:`~.message_ix.make_df`.
        If the value is :class:`float`, it overwrites the "value" column; if
        :class:`pint.Quantity`, its magnitude overwrites "value" and its units the
        "units" column, as a formatted string.

    Returns
    -------
    :class:`dict` of :class:`pandas.DataFrame`
        one for each parameter in `par_values`.

    Examples
    --------
    >>> input = make_df("input", ...)
    >>> cf_tl = make_matched_dfs(
    >>>     input,
    >>>     capacity_factor=1,
    >>>     technical_lifetime=pint.Quantity(8, "year"),
    >>> )
    """
    data = ChainMap(dict(), base)
    result = {}
    for par, value in par_value.items():
        data.maps[0] = (
            dict(value=value.magnitude, unit=f"{value.units:~}")
            if isinstance(value, pint.Quantity)
            else dict(value=value)
        )
        result[par] = (
            message_ix.make_df(par, **data).drop_duplicates().reset_index(drop=True)
        )
    return result


def make_source_tech(
    info: Union[message_ix.Scenario, ScenarioInfo], common, **values
) -> Dict[str, pd.DataFrame]:
    """Return parameter data for a ‘source’ technology.

    The technology has no inputs; its output commodity and/or level are determined by
    `common`; either single values, or :obj:`None` if the result will be
    :meth:`~pandas.DataFrame.pipe`'d through :func:`broadcast`.

    Parameters
    ----------
    info : Scenario or ScenarioInfo
    common : dict
        Passed to :func:`~message_ix.make_df`.
    **values
        Values for 'capacity_factor' (optional; default 1.0), 'output', 'var_cost', and
        optionally 'technical_lifetime'.

    Returns
    -------
    dict
        Suitable for :func:`add_par_data`.
    """
    # Check arguments
    if isinstance(info, message_ix.Scenario):
        info = ScenarioInfo(info)

    values.setdefault("capacity_factor", 1.0)
    missing = {"capacity_factor", "output", "var_cost"} - set(values.keys())
    if len(missing):
        raise ValueError(f"make_source_tech() needs values for {repr(missing)}")
    elif "technical_lifetime" not in values:
        log.debug("No technical_lifetime for source technology")

    # Create data for "output"
    result = dict(
        output=message_ix.make_df(
            "output",
            value=values.pop("output"),
            year_act=info.Y,
            year_vtg=info.Y,
            **common,
        )
        .pipe(broadcast, node_loc=nodes_ex_world(info.N))
        .pipe(same_node)
    )

    # Add data for other parameters
    result.update(make_matched_dfs(base=result["output"], **values))

    return result


def maybe_query(series: pd.Series, query: Optional[str]) -> pd.Series:
    """Apply :meth:`pandas.DataFrame.query` if the `query` arg is not :obj:`None`.

    :meth:`~pandas.DataFrame.query` is not chainable (`pandas-dev/pandas#37941
    <https://github.com/pandas-dev/pandas/issues/37941>`_). Use this function with
    :meth:`pandas.Series.pipe`, passing an argument that may be :obj:`None`, to have a
    chainable query operation that can be a no-op.
    """
    # Convert Series to DataFrame, query(), then retrieve the single column
    return series if query is None else series.to_frame().query(query)[0]


def merge_data(
    base: MutableMapping[str, pd.DataFrame], *others: Mapping[str, pd.DataFrame]
) -> None:
    """Merge dictionaries of DataFrames together into `base`."""
    for other in others:
        for par, df in other.items():
            base[par] = pd.concat([base.get(par, None), df])


def minimum_version(expr: str) -> Callable:
    """Decorator for functions that require a minimum version of some upstream package.

    See :func:`.prepare_reporter` / :func:`.test_prepare_reporter` for a usage example.

    Parameters
    ----------
    expr :
        Like "example 1.2.3.post0". The condition for the decorated function is that
        the installed version must be equal to or greater than this version.
    """

    from packaging.version import parse

    package, v_min = expr.split(" ")
    v_package = version(package)
    condition = parse(v_package) < parse(v_min)
    message = f" with {package} {v_package} < {v_min}"

    # Create the decorator
    def decorator(func):
        name = f"{func.__module__}.{func.__name__}()"

        # Wrap `func`
        def wrapper(*args, **kwargs):
            if condition:
                raise NotImplementedError(f"{name}{message}.")
            return func(*args, **kwargs)

        update_wrapper(wrapper, func)

        # Create a test function decorator
        def marker(test_func):
            # Import pytest only when there is a test function to mark
            import pytest

            # Create the mark
            mark = pytest.mark.xfail(
                condition=condition,
                raises=NotImplementedError,
                reason=f"Not supported{message}",
            )

            # Attach to the test function
            try:
                test_func.pytestmark.append(mark)
            except AttributeError:
                test_func.pytestmark = [mark]

            return test_func

        # Store the decorator on the wrapped function
        setattr(wrapper, "minimum_version", marker)

        return wrapper

    return decorator


def replace_par_data(
    scenario: message_ix.Scenario,
    parameters: Union[str, Sequence[str]],
    filters: Mapping[str, Union[str, int, Collection[str], Collection[int]]],
    to_replace: Mapping[str, Union[Mapping[str, str], Mapping[int, int]]],
) -> None:
    """Replace data in `parameters` of `scenario`.

    Parameters
    ----------
    scenario
        Scenario in which to replace data.
    parameters : str or sequence of str
        Name(s) of parameters in which to replace data.
    filters
        Passed to :meth:`.Scenario.par` argument of the same name.
    to_replace
        Passed to :meth:`pandas.DataFrame.replace` argument of the same name.

    Examples
    --------
    Replace data in the "relation_activity" parameter for a particular technology and
    relation: assign the same values as entries in a different relation name for the
    same technology.

    >>> replace_par_data(
    ...     scenario,
    ...     "relation_activity",
    ...     dict(technology="hp_gas_i", relation="CO2_r_c"),
    ...     dict(relation={"CO2_r_c": "CO2_ind"}),
    ... )

    """
    from message_ix_models.model.build import apply_spec

    pars = parameters.split() if isinstance(parameters, str) else parameters

    # Create a Spec that requires `scenario` to have all the set elements mentioned by
    # `filters` and/or `replacements`
    s = Spec()
    for k, v in filters.items():
        s.require.set[k].extend([v] if isinstance(v, (str, int)) else v)
    for k, v in to_replace.items():
        s.require.set[k].extend(v.keys())
        s.require.set[k].extend(v.values())

    # Use apply_spec() simply to check that `scenario` contains the expected items,
    # before attempting to modify data
    apply_spec(scenario, s)

    msg = f"Replace {filters!r} with {to_replace!r}"
    log.info(msg)
    for par_name in pars:
        with scenario.transact(f"{msg} in {par_name!r}"):
            # Base data, to be replaced
            to_remove = scenario.par(par_name, filters=filters)
            # Remove the base data
            scenario.remove_par(par_name, to_remove.drop(columns=["value", "unit"]))
            # Add the modified data
            scenario.add_par(par_name, to_remove.replace(to_replace))

            log.info(f"{len(to_remove)} obs in {par_name!r}")


def same_node(df: pd.DataFrame, from_col="node_loc") -> pd.DataFrame:
    """Fill 'node_{,dest,loc,origin,rel,share}' in `df` from `from_col`."""
    cols = list(
        set(df.columns)
        & ({"node", "node_loc", "node_origin", "node_dest", "node_rel", "node_share"})
        - {from_col}
    )
    return df.assign(**{c: copy_column(from_col) for c in cols})


def same_time(df: pd.DataFrame) -> pd.DataFrame:
    """Fill 'time_origin'/'time_dest' in `df` from 'time'."""
    cols = list(set(df.columns) & {"time_origin", "time_dest"})
    return df.assign(**{c: copy_column("time") for c in cols})


def show_versions() -> str:
    """Output of :func:`ixmp.show_versions`, as a :class:`str`."""
    from io import StringIO

    from . import ixmp
    from ._logging import preserve_log_handlers

    # Retrieve package versions
    buf = StringIO()

    # show_versions() imports pyam-iamc, which in turn imports ixmp4, which removes all
    # handlers from the root logger (?!). Preserve the message-ix-models logging config.
    with preserve_log_handlers():
        ixmp.show_versions(buf)

    return buf.getvalue()


# FIXME Reduce complexity from 14 to ≤13
def strip_par_data(  # noqa: C901
    scenario: message_ix.Scenario,
    set_name: str,
    element: str,
    dry_run: bool = False,
    dump: Optional[Dict[str, pd.DataFrame]] = None,
) -> int:
    """Remove `element` from `set_name` in scenario, optionally dumping to `dump`.

    Parameters
    ----------
    dry_run : bool, *optional*
        If :data:`True`, only show what would be done.
    dump : dict, *optional*
        If provided, stripped data are stored in this dictionary. Otherwise, they are
        discarded.

    Returns
    -------
    int
       Total number of rows removed across all parameters.

    See also
    --------
    add_par_data
    """
    par_list = scenario.par_list()
    no_data = set()  # Names of parameters with no data being stripped
    total = 0  # Total observations stripped

    if dump is None:
        pars = []  # Don't iterate over parameters unless dumping
    else:
        log.info(
            f"Remove data with {set_name}={element!r}"
            + (" (DRY RUN)" if dry_run else "")
        )
        # Iterate over parameters with ≥1 dimensions indexed by `set_name`
        pars = iter_parameters(set_name, scenario=scenario)

    for par_name in pars:
        if par_name not in par_list:  # pragma: no cover
            log.warning(
                f"  MESSAGEix parameter {par_name!r} missing in Scenario {scenario.url}"
            )
            continue

        # Iterate over dimensions indexed by `set_name`
        for dim, _ in filter(
            lambda item: item[1] == set_name,
            zip(scenario.idx_names(par_name), scenario.idx_sets(par_name)),
        ):
            # Check for contents of par_name that include `element`
            par_data = scenario.par(par_name, filters={dim: element})
            N = len(par_data)
            total += N

            if N == 0:
                # No data; no need to do anything further
                no_data.add(par_name)
                continue
            elif dump is not None:
                dump[par_name] = pd.concat(
                    [dump.get(par_name, pd.DataFrame()), par_data]
                )

            log.info(f"  {N} rows in {par_name!r}")

            # Show some debug info
            for col in filter(
                lambda c: c != set_name and c in par_data.columns,
                ("commodity", "level", "technology"),
            ):
                log.info(f"  with {col}={sorted(par_data[col].unique())}")

            if dry_run:
                continue

            # Actually remove the data
            scenario.remove_par(par_name, key=par_data)

            # NB would prefer to do the following, but raises an exception:
            # scenario.remove_par(par_name, key={set_name: [value]})

    if not dry_run and dump is not None:
        log.info(f"  {total} rows total")
    if no_data:
        log.debug(f"No data removed from {len(no_data)} other parameters")

    if not dry_run:
        log.info(f"Remove {element!r} from set {set_name!r}")
        try:
            scenario.remove_set(set_name, element)
        except Exception as e:
            if "does not have an element" in str(e):
                log.info("  …not found")
            else:  # pragma: no cover
                raise

    return total
